import contextlib
import decimal
import functools
import itertools
import json
import operator
import pathlib
import sys
import time
import uuid
import warnings

import attr
import canmatrix
import sunspecdemo.demos
import sunspecdemo.utils
import sunspec.core.client
import twisted.internet.defer

import epyqlib.busproxy
import epyqlib.canneo
import epyqlib.device
import epyqlib.nv
import epyqlib.utils.qt
import epyqlib.utils.twisted
import epyqlib.utils.units
import epyqlib.updateepc


class FormatVersionError(Exception):
    pass


class AlreadyLoadedError(Exception):
    pass


class BusAlreadySetError(Exception):
    pass


class RestartTimeoutError(Exception):
    pass


supported_version = [2]


def format_version_validator(instance, attribute, value):
    if value != supported_version:
        raise FormatVersionError(
            "Only format_version {} is supported".format(
                ".".join(str(v) for v in supported_version),
            ),
        )


def parse_units(unit_string):
    if unit_string is not None:
        unit_string = unit_string.replace("%", " percent ")
        # TODO: this is terrible T1179
        unit_string = unit_string.replace("nominal", "")
        unit_string = unit_string.replace("Power", "")
        unit_string = unit_string.replace("cos()", "")

    return epyqlib.utils.units.registry.parse_units(unit_string)


@attr.s
class Definition:
    base_path = attr.ib()
    format_version = attr.ib(
        validator=format_version_validator,
    )
    can_path = attr.ib(converter=pathlib.Path)
    # can_configuration = attr.ib()
    nv_configuration = attr.ib()
    node_id_type = attr.ib()
    access_level_path = attr.ib()
    access_password_path = attr.ib()
    node_id = attr.ib(default=247)
    controller_id = attr.ib(default=65)
    # nv_range_check_overridable = attr.ib()
    # node_id_type = attr.ib()
    # name = attr.ib()
    # tabs = attr.ib()
    # ui_paths = attr.ib()
    # module = attr.ib()
    # parameter_hierarchy = attr.ib()
    # nv_meta_enum = attr.ib()

    def load_can(self):
        (matrix,) = canmatrix.formats.loadp(
            str(self.base_path / self.can_path),
            symImportEncoding="utf-8",
        ).values()

        return matrix

    @classmethod
    def load(cls, file, base_path=None):
        if base_path is None:
            base_path = pathlib.Path(file.name).parents[0]

        return cls.loads(s=file.read(), base_path=base_path)

    @classmethod
    def loadp(cls, path):
        with open(path) as f:
            return cls.load(f)

    @classmethod
    def loads(cls, s, base_path):
        raw = json.loads(s)

        access_level_path = raw.get("access_level_path")
        if access_level_path is not None:
            access_level_path = access_level_path.split(";")

        access_password_path = raw.get("access_password_path")
        if access_password_path is not None:
            access_password_path = access_password_path.split(";")

        return Definition(
            base_path=base_path,
            format_version=raw["format_version"],
            can_path=base_path / raw["can_path"],
            nv_configuration=raw["nv_configuration"],
            node_id=raw.get("node_id"),
            node_id_type=raw.get(
                "node_id_type",
                next(iter(epyqlib.device.node_id_types)),
            ),
            controller_id=raw.get("controller_id"),
            access_level_path=access_level_path,
            access_password_path=access_password_path,
        )


operator_map = {
    "<": operator.lt,
    "<=": operator.le,
    "==": operator.eq,
    "!=": operator.ne,
    ">=": operator.ge,
    ">": operator.gt,
}

reverse_operator_map = {v: k for k, v in operator_map.items()}


@attr.s
class Signal:
    signal = attr.ib()
    device = attr.ib()

    async def _get_raw(self, stale_after=0.1, timeout=1):
        # TODO: uh...  why not time.monotonic()?
        start = time.time()

        last_received = self.signal.last_received()

        if last_received is None or last_received <= start - stale_after:
            await epyqlib.utils.qt.signal_as_deferred(
                self.signal.value_set,
                timeout=timeout,
            )

        return self.signal.value

    async def get(self, stale_after=0.1, timeout=1, enumeration_as_string=True):
        await self._get_raw(stale_after=stale_after, timeout=timeout)

        return self._to_human(enumeration_as_string=enumeration_as_string)

    def _to_human(self, enumeration_as_string):
        value = self.signal.to_human(value=self.signal.value)

        if enumeration_as_string and value in self.signal.enumeration:
            return self.signal.enumeration[value]

        return value * self.units()

    def set(self, value):
        units = self.units()
        if units != epyqlib.utils.units.registry.dimensionless:
            value = value.to(units).magnitude

        self.signal.set_human_value(value)
        self.signal.frame._send(update=True)

    # TODO: get rid of this and update everything to have .set() itself
    #       be an async
    async def async_set(self, value):
        return self.set(value=value)

    def units(self):
        return parse_units(self.signal.unit)

    def cyclic_send(self, period):
        self.device.cyclic_send_signal(self, period=period)

    async def wait_for(self, op, value, timeout):
        op = operator_map.get(op, op)
        operator_string = reverse_operator_map.get(op, str(op))

        async def check():
            present_value = await self.get()
            return op(present_value, value)

        await epyqlib.utils.twisted.wait_for(
            check=check,
            timeout=timeout,
            message=(
                f"{self.signal.name} not {operator_string} {value} "
                f"within {timeout:.1f} seconds"
            ),
        )

    def scaling_factor(self):
        return self.signal.factor

    def decimal_places(self):
        return self.signal.get_decimal_places()

    def f_string(self):
        return f".{self.decimal_places()}f"

    @contextlib.asynccontextmanager
    async def temporary_set(
        self,
        value=None,
        read_context=None,
        set_context=None,
        restoration_context=None,
    ):
        @contextlib.asynccontextmanager
        async def async_null_context(enter_result=None):
            with contextlib.nullcontext(enter_result=enter_result) as result:
                yield result

        if set_context is None:
            set_context = async_null_context
        if restoration_context is None:
            restoration_context = async_null_context

        original = self._to_human(enumeration_as_string=True)

        try:
            async with set_context():
                self.set(value=value)

            yield
        finally:
            if original is not None:
                async with restoration_context():
                    self.set(value=original)


@attr.s
class Nv:
    nv = attr.ib()
    device = attr.ib()

    async def set(
        self,
        value=None,
        user_default=None,
        factory_default=None,
        minimum=None,
        maximum=None,
    ):
        values = {
            epyqlib.nv.MetaEnum.maximum: maximum,
            epyqlib.nv.MetaEnum.minimum: minimum,
            epyqlib.nv.MetaEnum.factory_default: factory_default,
            epyqlib.nv.MetaEnum.user_default: user_default,
            epyqlib.nv.MetaEnum.value: value,
        }

        values = {meta: value for meta, value in values.items() if value is not None}

        for meta, value in values.items():
            await self.set_meta(value=value, meta=meta)

    async def set_meta(self, value, meta):
        units = self.units()
        if units != epyqlib.utils.units.registry.dimensionless:
            value = value.to(units).magnitude

        if meta == epyqlib.nv.MetaEnum.value:
            self.nv.set_human_value(value)
        else:
            getattr(self.nv.meta, meta.name).set_human_value(value)

        # TODO: verify value was accepted
        await self.device.nvs.protocol.write(
            nv_signal=self.nv,
            meta=meta,
        )

    @contextlib.asynccontextmanager
    async def temporary_set(
        self,
        value=None,
        user_default=None,
        factory_default=None,
        minimum=None,
        maximum=None,
        read_context=None,
        set_context=None,
        restoration_context=None,
    ):
        @contextlib.asynccontextmanager
        async def async_null_context(enter_result=None):
            with contextlib.nullcontext(enter_result=enter_result) as result:
                yield result

        if read_context is None:
            read_context = async_null_context
        if set_context is None:
            set_context = async_null_context
        if restoration_context is None:
            restoration_context = async_null_context

        values = {
            epyqlib.nv.MetaEnum.maximum: maximum,
            epyqlib.nv.MetaEnum.minimum: minimum,
            epyqlib.nv.MetaEnum.factory_default: factory_default,
            epyqlib.nv.MetaEnum.user_default: user_default,
            epyqlib.nv.MetaEnum.value: value,
        }

        values = {meta: value for meta, value in values.items() if value is not None}

        original = {}

        try:
            async with read_context():
                for meta, value in values.items():
                    original[meta] = await self.get(meta=meta)

            async with set_context():
                for meta, value in values.items():
                    await self.set_meta(value=value, meta=meta)

            yield
        finally:
            async with restoration_context():
                for meta, value in original.items():
                    await self.set_meta(value=value, meta=meta)

    async def get(self, meta=epyqlib.nv.MetaEnum.value):
        value, _meta = await self.device.nvs.protocol.read(
            nv_signal=self.nv,
            meta=meta,
        )

        return value * self.units()

    def units(self):
        return parse_units(unit_string=self.nv.unit)

    async def wait_for(self, op, value, timeout, ignore_read_failures=False):
        op = operator_map.get(op, op)
        operator_string = reverse_operator_map.get(op, str(op))

        async def check():
            try:
                own_value = await self.get()
            except epyqlib.twisted.nvs.RequestTimeoutError:
                if ignore_read_failures:
                    return False

                raise

            return op(own_value, value)

        await epyqlib.utils.twisted.wait_for(
            check=check,
            timeout=timeout,
            message=(
                f"{self.nv.name} not {operator_string} {value} "
                f"within {timeout:.1f} seconds"
            ),
        )

    def parameter_uuid(self):
        return self.nv.parameter_uuid


@attr.s
class Device:
    definition_path = attr.ib(converter=pathlib.Path)
    definition = attr.ib(default=None)
    canmatrix = attr.ib(default=None)
    neo = attr.ib(default=None)
    nvs = attr.ib(default=None)
    bus: epyqlib.busproxy.BusProxy = attr.ib(default=None)
    cyclic_frames = attr.ib(default=attr.Factory(set))
    default_elevated_access_level = attr.ib(default=None)
    default_access_level_password = attr.ib(default=None)
    save_nv = attr.ib(default=None)
    save_nv_value = attr.ib(default=None)
    uuid = attr.ib(default=uuid.uuid4)

    def load(self):
        if self.definition is not None:
            raise AlreadyLoadedError("The definition has already been loaded")

        with epyqlib.updateepc.updated(self.definition_path) as updated:
            self.definition = Definition.loadp(updated)
            matrix = self.definition.load_can()

        node_id_adjust = functools.partial(
            epyqlib.device.node_id_types[self.definition.node_id_type],
            device_id=self.definition.node_id,
            controller_id=self.definition.controller_id,
        )

        self.neo = epyqlib.canneo.Neo(
            matrix=matrix,
            node_id_adjust=node_id_adjust,
        )

        nv_neo = epyqlib.canneo.Neo(
            matrix=matrix,
            frame_class=epyqlib.nv.Frame,
            signal_class=epyqlib.nv.Nv,
            strip_summary=False,
            node_id_adjust=node_id_adjust,
        )
        self.nvs = epyqlib.nv.Nvs(
            neo=nv_neo,
            configuration=self.definition.nv_configuration,
            access_level_path=self.definition.access_level_path,
            access_password_path=self.definition.access_password_path,
        )

        self.save_nv = self.nv(
            self.nvs.save_frame.mux_name,
            self.nvs.save_signal.name,
        )
        self.save_nv_value = self.nvs.save_value

    def set_bus(self, bus):
        if self.bus is not None:
            raise BusAlreadySetError()

        try:
            self.neo.set_bus(bus=bus)
            self.nvs.set_bus(bus=bus)
        except:
            # TODO: actually rollback a partial setting
            self.bus = object()
            raise

        self.bus = bus
        self.bus.notifier.add(self.neo)
        # TODO: really think through what is proper...  won't this keep the
        #       nv objects from getting updated?
        # self.bus.notifier.add(self.nvs)

    # @functools.lru_cache(maxsize=512)
    def signal(self, *path):
        return Signal(
            signal=self.neo.signal_by_path(*path),
            device=self,
        )

    def signal_from_uuid(self, uuid_):
        return Signal(
            signal=self.neo.signal_from_uuid[uuid_],
            device=self,
        )

    # @functools.lru_cache(maxsize=512)
    def nv(self, *path):
        return Nv(
            nv=self.nvs.signal_from_names(*path),
            device=self,
        )

    def nv_from_uuid(self, uuid_):
        return Nv(
            nv=self.nvs.nv_from_uuid[uuid_],
            device=self,
        )

    def parameter_from_uuid(self, uuid_):
        try:
            return self.nv_from_uuid(uuid_=uuid_)
        except KeyError:
            return self.signal_from_uuid(uuid_=uuid_)

    async def active_to_nv(self, wait=False):
        # TODO: dedupe 8795477695t46542676781543768139
        await self.save_nv.set(value=self.save_nv_value)

        if wait:
            await self.wait_for_nv_save_completion()

    async def wait_for_nv_save_completion(self):
        nv = self.nv("StatusWarnings", "eeSaveInProgress")

        await epyqlib.utils.twisted.sleep(2)

        await nv.wait_for(
            op="==",
            value=0,
            timeout=120,
            ignore_read_failures=True,
        )

    def cyclic_send_signal(self, signal, period):
        frame = signal.signal.frame
        frame.cyclic_request(self.uuid, period)
        if period is not None:
            self.cyclic_frames.add(frame)
        else:
            self.cyclic_frames.discard(frame)

    def cancel_all_cyclic_sends(self):
        for frame in self.cyclic_frames:
            frame.cyclic_request(self.uuid, None)

    async def get_access_level(self):
        nv = Nv(nv=self.nvs.access_level_node, device=self)
        access_level = await nv.get()
        return access_level

    async def get_check_limits(self):
        nv = self.nv_from_uuid(
            uuid.UUID("bd7c3c96-bde9-4b4b-a646-e1d06a7cc24f"),
        )
        value = await nv.get()

        return value

    async def get_password(self):
        nv = self.nv_from_uuid(
            uuid.UUID("cc438574-bec0-4443-8a25-785e41240c1b"),
        )
        value = await nv.get()

        return value

    async def set_access_level(self, level=None, password=None, check_limits=True):
        if level is None:
            level = self.default_elevated_access_level

        if password is None:
            password = self.default_access_level_password

        self.nvs.password_node.set_value(password)
        self.nvs.access_level_node.set_value(level)
        check_limits_nv = self.nv_from_uuid(
            uuid.UUID("bd7c3c96-bde9-4b4b-a646-e1d06a7cc24f"),
        )
        check_limits_nv.nv.set_value(check_limits)

        selected_nodes = tuple(
            node
            for node in (
                self.nvs.password_node,
                self.nvs.access_level_node,
                check_limits_nv.nv,
            )
            if node is not None
        )

        await self.nvs.write_all_to_device(
            only_these=selected_nodes,
            meta=[epyqlib.nv.MetaEnum.value],
        )

    @contextlib.asynccontextmanager
    async def temporary_access_level(
        self,
        level=None,
        password=None,
        check_limits=True,
    ):
        access_level_parameter = self.nv(*self.definition.access_level_path[1:])
        original_access_level = await access_level_parameter.get()

        check_limits_nv = self.nv_from_uuid(
            uuid.UUID("bd7c3c96-bde9-4b4b-a646-e1d06a7cc24f"),
        )
        original_check_limits = await check_limits_nv.get()

        try:
            await self.set_access_level(
                level=level,
                password=password,
                check_limits=check_limits,
            )
            yield
        finally:
            try:
                await self.set_access_level(
                    level=level,
                    password=password,
                    check_limits=original_check_limits,
                )
            finally:
                await self.set_access_level(
                    level=original_access_level,
                    password=password,
                )

    async def reset(self, timeout=20, sleep=5):
        # SoftwareReset:InitiateReset
        reset_parameter = self.parameter_from_uuid(
            uuid_=uuid.UUID("b582085d-7734-4260-ab97-47e50a41b06c"),
        )

        # TODO: just accept the 1s or whatever default timeout?  A set without
        #       waiting for the response could be nice.  (or embedded sending
        #       a response)

        with contextlib.suppress(epyqlib.twisted.nvs.RequestTimeoutError):
            await reset_parameter.set(value=1)

        self.bus.transmit = False
        self.bus.reset()

        if sleep > 0:
            await epyqlib.utils.twisted.sleep(sleep)

        await self.reconnect(timeout=timeout)

    async def reconnect(self, timeout=5):
        # Serial Number
        a_parameter_that_can_be_read = self.parameter_from_uuid(
            uuid_=uuid.UUID("390f27ea-6f28-4313-b183-5f37d007ccd1"),
        )

        start = time.monotonic()
        end = start + timeout
        for retry in itertools.count():
            self.bus.transmit = False
            self.bus.reset()
            self.bus.reconnect()
            await epyqlib.utils.twisted.sleep(0.500)
            self.bus.reset()
            self.bus.transmit = True

            try:
                for _ in range(5):
                    await epyqlib.utils.twisted.sleep(0.2)
                    await a_parameter_that_can_be_read.get()
            except (
                epyqlib.twisted.nvs.RequestTimeoutError,
                epyqlib.twisted.nvs.SendFailedError,
            ) as e:
                now = time.monotonic()
                if now > end:
                    raise RestartTimeoutError(
                        f"Failed on retry {retry} after {now - start} seconds"
                    ) from e
                continue
            else:
                break

    async def wait_through_power_on_reset(self):
        status_signal = self.signal_from_uuid(
            uuid_=uuid.UUID("6392782a-b886-45a0-9642-dd4f47cd2a59"),
        )

        await status_signal.wait_for(
            op="!=",
            # TODO: stop comparing strings...
            value="Power On Reset",
            timeout=60,
        )

    async def to_nv(self):
        # TODO: dedupe 8795477695t46542676781543768139
        await self.nvs.module_to_nv()

    async def get_serial_number(self):
        nv = self.nv_from_uuid(
            uuid_=uuid.UUID("390f27ea-6f28-4313-b183-5f37d007ccd1"),
        )
        value = await nv.get()
        return value

    async def clear_faults(self):
        clear_faults_signal = self.signal_from_uuid(
            uuid_=uuid.UUID("62b6dc82-c93a-454a-a643-dd8a7b2a220e"),
        )
        clear_faults_status_signal = self.signal_from_uuid(
            uuid_=uuid.UUID("d84e5184-696d-487c-8850-cc904a7c018f"),
        )

        clear_faults_signal.set(value=False)
        await clear_faults_status_signal.wait_for(
            op="==",
            value="Normal",
            timeout=1,
        )

        clear_faults_signal.set(value=True)
        await clear_faults_status_signal.wait_for(
            op="==",
            value="Clear Faults",
            timeout=1,
        )

        clear_faults_signal.set(value=False)
        await clear_faults_status_signal.wait_for(
            op="==",
            value="Normal",
            timeout=1,
        )


async def set_key_inplace(
    key_nvs,
    tag_nvs,
    existing_key,
    replacement_key,
    serial_number,
):
    new_level_key = attr.evolve(existing_key, level=replacement_key)
    new_low_key = attr.evolve(
        new_level_key,
        value=new_level_key.value.evolve_low(replacement_key.value.low_32()),
    )

    to_write = [
        (key_nvs.access_level, existing_key, replacement_key.level),
        (key_nvs.low, new_level_key, replacement_key.value.low_32()),
        (key_nvs.high, new_low_key, replacement_key.value.high_32()),
    ]

    for nv, present_key, value in to_write:
        tag = epyqlib.authorization.create_tag(
            key_value=present_key.value,
            serial_number=serial_number,
            parameter_uuid=nv.parameter_uuid(),
            meta_index=0,
            value=int(
                nv.nv.pack_bitstring(value=value),
                2,
            ),
        )

        async with temporary_set(
            nvs_and_values=(
                (tag_nvs.low, tag.value.low_32()),
                (tag_nvs.high, tag.value.high_32()),
            )
        ):
            await nv.set(value)
            print()


@contextlib.asynccontextmanager
async def temporary_key(
    key_nvs,
    tag_nvs,
    existing_key,
    replacement_key,
    serial_number,
):
    try:
        await set_key_inplace(
            key_nvs=key_nvs,
            tag_nvs=tag_nvs,
            existing_key=existing_key,
            replacement_key=replacement_key,
            serial_number=serial_number,
        )

        yield
    finally:
        await set_key_inplace(
            key_nvs=key_nvs,
            tag_nvs=tag_nvs,
            existing_key=replacement_key,
            replacement_key=existing_key,
            serial_number=serial_number,
        )


@attr.s(frozen=True)
class KeyNvs:
    low = attr.ib()
    high = attr.ib()
    access_level = attr.ib()
    index = attr.ib()

    @classmethod
    def from_index(cls, index, device):
        # These UUIDs are for key 0

        return cls(
            low=device.nv_from_uuid(
                uuid_=uuid.UUID("f046a08b-9308-4da7-ae5e-b931d610b3f4"),
            ),
            high=device.nv_from_uuid(
                uuid_=uuid.UUID("3e021694-d565-4636-8a5d-3e0976d1112c"),
            ),
            access_level=device.nv_from_uuid(
                uuid_=uuid.UUID("07c45cff-0b95-413d-9aed-01a4f514e8da"),
            ),
            index=index,
        )


@attr.s(frozen=True)
class TagNvs:
    low = attr.ib()
    high = attr.ib()
    index = attr.ib()

    @classmethod
    def from_index(cls, index, device):
        return cls(
            low=device.nv(f"AuthTag{index}Low32", "Low32"),
            high=device.nv(f"AuthTag{index}High32", "High32"),
            index=index,
        )


@contextlib.asynccontextmanager
async def temporary_set(nvs_and_values):
    async with contextlib.AsyncExitStack() as exit_stack:
        for nv, value in nvs_and_values:
            await exit_stack.enter_async_context(nv.temporary_set(value=value))

        yield


@attr.s
class AccessLevel:
    name = attr.ib()
    level = attr.ib()
    password = attr.ib()


@attr.s
class SunSpecNv:
    nv = attr.ib()
    model = attr.ib()
    # device = attr.ib()

    async def set(self, value):
        units = self.units()
        if units != epyqlib.utils.units.registry.dimensionless:
            value = value.to(units).magnitude

        self.nv.value = value
        self.nv.write()

    @contextlib.asynccontextmanager
    async def temporary_set(
        self,
        value,
        read_context=None,
        set_context=None,
        restoration_context=None,
    ):
        @contextlib.asynccontextmanager
        async def async_null_context(enter_result=None):
            with contextlib.nullcontext(enter_result=enter_result) as result:
                yield result

        if read_context is None:
            read_context = async_null_context
        if set_context is None:
            set_context = async_null_context
        if restoration_context is None:
            restoration_context = async_null_context

        original = None

        try:
            async with read_context():
                original = await self.get()

            async with set_context():
                await self.set(value=value)

            yield
        finally:
            if original is not None:
                async with restoration_context():
                    await self.set(value=original)

    async def get(self):
        self.model.read_points()

        value = decimal.Decimal(self.nv.value)
        scale_factor = self.nv.value_sf
        if scale_factor is None:
            scale_factor = 0
        value = round(value, -scale_factor)

        return value * self.units()

    def units(self):
        return parse_units(unit_string=self.nv.point_type.units)


@attr.s
class SunSpecDevice:
    model_path = attr.ib(converter=pathlib.Path)
    device = attr.ib(default=None)
    cyclic_frames = attr.ib(default=attr.Factory(set))
    default_elevated_access_level = attr.ib(default=None)
    default_access_level_password = attr.ib(default=None)
    save_nv = attr.ib(default=None)
    save_nv_value = attr.ib(default=None)
    uuid_to_point = attr.ib(default=None)
    uuid_to_model = attr.ib(default=None)
    uuid = attr.ib(default=uuid.uuid4)

    def load(
        self,
        slave_id=1,
        device_type=sunspec.core.client.RTU,
        name="/dev/ttyUSB0",
        baudrate=115200,
        timeout=1,
        parity=sunspec.core.client.PARITY_NONE,
        ipaddr=None,
        ipport=502,
        pathlist=None,
        trace=False,
    ):
        message = (
            "SunSpecDevice.load() method is deprecated and will be"
            + " removed in the future. Use SunSpecDevice.load_rtu()"
            + " or SunSpecDevice.load_tcp() instead."
        )
        warnings.warn(message=message, category=DeprecationWarning)

        with sunspecdemo.utils.fresh_smdx_path(self.model_path):
            self.device = sunspec.core.client.SunSpecClientDevice(
                slave_id=slave_id,
                device_type=device_type,
                name=name,
                baudrate=baudrate,
                timeout=timeout,
                parity=parity,
                ipaddr=ipaddr,
                ipport=ipport,
                pathlist=pathlist,
                trace=trace,
            )

    def load_rtu(
        self,
        slave_id=1,
        name="/dev/ttyUSB0",
        baudrate=115200,
        timeout=1,
        parity=sunspec.core.client.PARITY_NONE,
    ):
        with sunspecdemo.utils.fresh_smdx_path(self.model_path):
            self.device = sunspec.core.client.SunSpecClientDevice(
                slave_id=slave_id,
                device_type=sunspec.core.client.RTU,
                name=name,
                baudrate=baudrate,
                timeout=timeout,
                parity=parity,
            )

    def load_tcp(
        self,
        address,
        slave_id=1,
        timeout=1,
        port=None,
    ):
        with sunspecdemo.utils.fresh_smdx_path(self.model_path):
            self.device = sunspec.core.client.SunSpecClientDevice(
                slave_id=slave_id,
                device_type=sunspec.core.client.TCP,
                timeout=timeout,
                ipaddr=address,
                ipport=port,
            )

    # def signal_from_uuid(self, uuid_) -> SunSpecNv:
    #     return self.nv_from_uuid(uuid_=uuid_)

    def nv_from_uuid(self, uuid_) -> SunSpecNv:
        return SunSpecNv(
            nv=self.uuid_to_point[uuid_],
            model=self.uuid_to_model[uuid_],
            # device=self,
        )

    # no 'signals' so just alias
    parameter_from_uuid = nv_from_uuid

    def map_uuids(self):
        def get_uuid(block, point):
            comment = point.point_type.notes

            for index in itertools.count():
                comment, uuid = epyqlib.canneo.strip_uuid_from_comment(
                    comment,
                )

                if uuid is None:
                    return uuid

                if block.type == "fixed":
                    return uuid

                if block.type == "repeating" and index == block.index - 1:
                    return uuid

        points = [
            [model, block, point]
            for model in self.device.device.models_list
            for block in model.blocks
            for point in [*block.points_list, *block.points_sf.values()]
            if point.point_type.notes is not None
        ]

        self.uuid_to_point = {
            get_uuid(block=block, point=point): point for model, block, point in points
        }

        self.uuid_to_model = {
            get_uuid(block=block, point=point): model for model, block, point in points
        }

    async def get_access_level(self):
        access_level_point = self.device.epc_control.model.points["AccLvl"]

        self.device.epc_control.read()

        return access_level_point.value

    async def get_check_limits(self):
        point = self.device.epc_control.model.points["ChkLmts"]
        self.device.epc_control.read()

        return point.value

    async def get_password(self):
        point = self.device.epc_control.model.points["Passwd"]
        self.device.epc_control.read()

        return point.value

    async def set_access_level(self, level=None, password=None, check_limits=True):
        if level is None:
            level = self.default_elevated_access_level

        if password is None:
            password = self.default_access_level_password

        access_level_point = self.device.epc_control.model.points["AccLvl"]
        password_point = self.device.epc_control.model.points["Passwd"]
        check_limits_point = self.device.epc_control.model.points["ChkLmts"]
        submit_point = self.device.epc_control.model.points["SubAccLvl"]

        sunspecdemo.demos.send_val(access_level_point, level)
        sunspecdemo.demos.send_val(password_point, password)
        sunspecdemo.demos.send_val(check_limits_point, check_limits)

        sunspecdemo.demos.send_val(submit_point, True)

    @contextlib.asynccontextmanager
    async def temporary_access_level(
        self,
        level=None,
        password=None,
        check_limits=True,
    ):
        check_limits_point = self.device.epc_control.model.points["ChkLmts"]

        original_access_level = await self.get_access_level()
        self.device.epc_control.read()
        original_check_limits = check_limits_point.value

        try:
            await self.set_access_level(
                level=level,
                password=password,
                check_limits=check_limits,
            )
            yield
        finally:
            await self.set_access_level(
                level=original_access_level,
                password=password,
                check_limits=original_check_limits,
            )

    async def reset(self, timeout=20, sleep=25):
        # SoftwareReset:InitiateReset
        reset_parameter = self.parameter_from_uuid(
            uuid_=uuid.UUID("b582085d-7734-4260-ab97-47e50a41b06c"),
        )

        # TODO: just accept the 1s or whatever default timeout?  A set without
        #       waiting for the response could be nice.  (or embedded sending
        #       a response)
        with contextlib.suppress(sunspec.core.client.SunSpecClientError):
            await reset_parameter.set(value=1)

        if sleep > 0:
            await epyqlib.utils.twisted.sleep(sleep)

        await self.reconnect(timeout=timeout)

    async def reconnect(self, timeout=20):
        # Serial Number
        a_parameter_that_can_be_read = self.parameter_from_uuid(
            uuid_=uuid.UUID("390f27ea-6f28-4313-b183-5f37d007ccd1"),
        )

        end = time.monotonic() + timeout
        while True:
            try:
                for _ in range(5):
                    await epyqlib.utils.twisted.sleep(0.2)
                    await a_parameter_that_can_be_read.get()
            except sunspec.core.client.SunSpecClientError as e:
                if time.monotonic() > end:
                    raise RestartTimeoutError() from e
                continue
            else:
                break

    async def to_nv(self, timeout=30):
        save_command_parameter = self.parameter_from_uuid(
            uuid.UUID("2c768acc-f88e-431c-8fc1-ea8d5b2ba253"),
        )
        save_in_progress_parameter = self.parameter_from_uuid(
            uuid.UUID("5d623539-a564-4374-b00d-492a0fbb2f55"),
        )

        await save_command_parameter.set(1)
        await epyqlib.utils.twisted.sleep(0.250)

        end = time.monotonic() + timeout
        while time.monotonic() < end:
            try:
                saving = await save_in_progress_parameter.get()
            except sunspec.core.client.SunSpecClientError:
                continue

            if not saving:
                break
        else:
            raise Exception()



import collections
import inspect
import itertools
import sys

import attr
import graham
import PyQt5.QtCore
import PyQt5.QtWidgets
import pytest
from pytestqt.qt_compat import qt_api

import epyqlib.attrsmodel
import epyqlib.tests.common
import epyqlib.searchbox
import epyqlib.treenode
import epyqlib.utils.general
import epyqlib.utils.qt

# See file COPYING in this source tree
__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


@graham.schemify(tag="parameter")
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Parameter(epyqlib.treenode.TreeNode):
    name = attr.ib(default="New Parameter")
    value = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return False

    can_delete = epyqlib.attrsmodel.childless_can_delete
    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


@graham.schemify(tag="group")
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@attr.s(hash=False)
class Group(epyqlib.treenode.TreeNode):
    name = attr.ib(default="New Group")
    children = attr.ib(
        default=attr.Factory(list),
        cmp=False,
        init=False,
        metadata={"valid_types": (Parameter, None)},
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()

    def can_drop_on(self, node):
        return isinstance(node, tuple(self.addable_types().values()))

    def can_delete(self, node=None):
        if node is None:
            return self.tree_parent.can_delete(node=self)

        return True

    remove_old_on_drop = epyqlib.attrsmodel.default_remove_old_on_drop
    child_from = epyqlib.attrsmodel.default_child_from
    internal_move = epyqlib.attrsmodel.default_internal_move
    check = epyqlib.attrsmodel.check_just_children


Root = epyqlib.attrsmodel.Root(
    default_name="Parameters", valid_types=(Parameter, Group)
)

types = epyqlib.attrsmodel.Types(
    types=(
        Root,
        Parameter,
        Group,
    ),
)

# TODO: CAMPid 943896754217967154269254167
def merge(name, *types):
    return tuple((x, name) for x in types)


columns = epyqlib.attrsmodel.columns(
    (
        (Parameter, "name"),
        (Group, "name"),
    ),
    ((Parameter, "value"),),
    merge("uuid", *types.types.values()),
)


TestAttrsModel = epyqlib.attrsmodel.build_tests(
    types=types,
    root_type=Root,
    columns=columns,
)


def make_a_model(
    root_cls=Root,
    group_cls=Group,
    parameter_cls=Parameter,
    columns=columns,
):
    root = root_cls(uuid="0f68fa9c-705d-4ba6-9ffa-406cb549a4dd")

    model = epyqlib.attrsmodel.Model(
        root=root,
        columns=columns,
    )

    group_a = group_cls(
        name="Group A",
        uuid="f5b7569e-9d7e-4433-a034-29c3e04d1ad4",
    )
    parameter_a_a = parameter_cls(
        name="Parameter A A",
        uuid="df286eb3-67f0-42d6-b56a-8ee1ded49248",
    )
    group_a_b = group_cls(
        name="Group A B",
        uuid="aee15e15-c5df-4e73-ae1a-9a5d4eaa798a",
    )
    parameter_b = parameter_cls(
        name="Parameter B",
        value=42,
        uuid="a1fd7abb-4760-472e-bc94-1ef4d2cfad62",
    )
    group_c = group_cls(
        name="Group C",
        uuid="2777e016-a3e6-470d-b20c-7a44904df710",
    )
    parameter_d = parameter_cls(
        name="Parameter D",
        value=42,
        uuid="97f1f9e1-5601-4304-a583-561df79c47be",
    )

    root.append_child(group_a)
    group_a.append_child(parameter_a_a)
    group_a.append_child(group_a_b)

    root.append_child(parameter_b)

    root.append_child(group_c)

    root.append_child(parameter_d)

    return model


def test_column_header_text(qtbot):
    model = make_a_model()

    assert model.model.horizontalHeaderItem(0).text() == "Name"
    assert model.model.horizontalHeaderItem(1).text() == "Value"


def test_model(qtmodeltester):
    model = make_a_model()

    qtmodeltester.check(model.model)


def test_search_column_0(qtbot):
    search_in_column(0, "Parameter B")


def test_search_column_2(qtbot):
    search_in_column(1, 42)


def search_in_column(column, target):
    model = make_a_model()

    proxy = PyQt5.QtCore.QSortFilterProxyModel()
    proxy.setSourceModel(model.model)

    view = PyQt5.QtWidgets.QTreeView()
    view.setModel(proxy)

    (index,) = proxy.match(
        proxy.index(0, column),
        PyQt5.QtCore.Qt.DisplayRole,
        target,
        1,
        PyQt5.QtCore.Qt.MatchRecursive,
    )


def test_proxy_search_column_0(qtbot):
    proxy_search_in_column(0, "Parameter B")


def test_proxy_search_column_0_child(qtbot):
    proxy_search_in_column(0, "Parameter A A")


def test_proxy_search_column_2(qtbot):
    proxy_search_in_column(1, 42)


def proxy_search_in_column(column, target):
    model = make_a_model()

    proxy = epyqlib.utils.qt.PySortFilterProxyModel(filter_column=0)
    proxy.setSourceModel(model.model)

    view = PyQt5.QtWidgets.QTreeView()
    view.setModel(proxy)

    (index,) = proxy.match(
        proxy.index(0, column),
        PyQt5.QtCore.Qt.DisplayRole,
        target,
        1,
        PyQt5.QtCore.Qt.MatchRecursive,
    )

    match_node = model.node_from_index(proxy.mapToSource(index))
    assert match_node is not None

    index = proxy.search(
        text=target,
        search_from=PyQt5.QtCore.QModelIndex(),
        column=column,
    )

    search_node = model.node_from_index(proxy.mapToSource(index))

    assert match_node is search_node


def node_from_name(model, name):
    (index,) = model.model.match(
        model.model.index(0, 0),
        PyQt5.QtCore.Qt.DisplayRole,
        name,
        1,
        PyQt5.QtCore.Qt.MatchRecursive,
    )

    return model.node_from_index(index)


@attr.s
class DataChangedCollector(PyQt5.QtCore.QObject):
    collected = PyQt5.QtCore.pyqtSignal("PyQt_PyObject")

    model = attr.ib()
    parameter = attr.ib()
    column = attr.ib()
    roles = attr.ib(default=(PyQt5.QtCore.Qt.DisplayRole,))

    def __attrs_post_init__(self):
        super().__init__()

    def collect(self, top_left, bottom_right, roles):
        # TODO: this is overly restrictive in that exactly the one cell
        #       must be changing rather than just being included in the
        #       range.

        right_one = all(
            (
                self.parameter is self.model.node_from_index(top_left),
                self.parameter is self.model.node_from_index(bottom_right),
                self.column == top_left.column() == bottom_right.column(),
                set(self.roles).issubset(roles),
            )
        )

        if right_one:
            parameter_name_index = top_left.siblingAtColumn(self.column)

            self.collected.emit(
                parameter_name_index.data(
                    PyQt5.QtCore.Qt.DisplayRole,
                )
            )


def test_data_changed(qtbot):
    model = make_a_model()

    parameter = node_from_name(model, "Parameter B")

    values = epyqlib.tests.common.Values(
        initial=42,
        input=[42, 37],
        expected=["37"],
    )

    parameter.value = values.initial

    data_changed = DataChangedCollector(
        model=model,
        parameter=parameter,
        column=1,
        roles=(PyQt5.QtCore.Qt.DisplayRole,),
    )

    model.model.dataChanged.connect(data_changed.collect)
    data_changed.collected.connect(values.collect)

    for value in values.input:
        parameter.value = value

    parameter.tree_parent.remove_child(child=parameter)

    parameter.value += 1

    assert tuple(values.collected) == tuple(values.expected)


def test_other_data_did_not_change(qtbot):
    model = make_a_model()

    parameter = node_from_name(model, "Parameter B")

    values = epyqlib.tests.common.Values(
        initial=42,
        input=[42, 37],
        expected=["37"],
    )

    parameter.value = values.initial

    data_changed = DataChangedCollector(
        model=model,
        parameter=parameter,
        column=1,
        roles=(PyQt5.QtCore.Qt.DisplayRole,),
    )

    other_parameter = node_from_name(model, "Parameter D")

    other_values = epyqlib.tests.common.Values(
        initial=12,
        input=[],
        expected=[],
    )

    other_parameter.value = other_values.initial

    other_data_changed = DataChangedCollector(
        model=model,
        parameter=other_parameter,
        column=1,
        roles=(PyQt5.QtCore.Qt.DisplayRole,),
    )

    model.model.dataChanged.connect(data_changed.collect)
    data_changed.collected.connect(values.collect)

    model.model.dataChanged.connect(other_data_changed.collect)
    other_data_changed.collected.connect(other_values.collect)

    for value in values.input:
        parameter.value = value

    parameter.tree_parent.remove_child(child=parameter)

    parameter.value += 1

    assert tuple(values.expected) == tuple(values.collected)
    assert tuple(other_values.expected) == tuple(other_values.collected)


def test_local_drag_n_drop(qtbot):
    model = make_a_model()
    model.add_drop_sources(model)

    parameter = node_from_name(model, "Parameter B")
    group = node_from_name(model, "Group C")

    values = epyqlib.tests.common.Values(
        initial=42,
        input=[42, 37, 23],
        expected=["37", "23"],
    )

    values_after_drop = epyqlib.tests.common.Values(
        initial=int(values.expected[-1]),
        input=[11, 13],
        expected=["11", "13"],
    )

    parameter.value = values.initial

    data_changed = DataChangedCollector(
        model=model,
        parameter=parameter,
        column=1,
        roles=(PyQt5.QtCore.Qt.DisplayRole,),
    )

    model.model.dataChanged.connect(data_changed.collect)
    data_changed.collected.connect(values.collect)

    for value in values.input:
        parameter.value = value

    assert tuple(values.expected) == tuple(values.collected)

    mime_data = model.mimeData((model.index_from_node(parameter),))

    model.dropMimeData(
        data=mime_data,
        action=PyQt5.QtCore.Qt.MoveAction,
        row=-1,
        column=0,
        parent=model.index_from_node(group),
    )

    data_changed.collected.connect(values_after_drop.collect)

    for value in values_after_drop.input:
        parameter.value = value

    parameter.tree_parent.remove_child(child=parameter)

    parameter.value += 1

    assert tuple(values_after_drop.expected) == tuple(values_after_drop.collected)


def test_prepopulated_connections(qtbot):
    values = epyqlib.tests.common.Values(
        initial=42,
        input=[42, 37, 23],
        expected=["37", "23"],
    )

    parameter = Parameter(
        name="Parameter A",
        value=values.initial,
    )

    root = Root()
    root.append_child(parameter)

    model = epyqlib.attrsmodel.Model(
        root=root,
        columns=columns,
    )

    data_changed = DataChangedCollector(
        model=model,
        parameter=parameter,
        column=1,
        roles=(PyQt5.QtCore.Qt.DisplayRole,),
    )

    model.model.dataChanged.connect(data_changed.collect)
    data_changed.collected.connect(values.collect)

    for value in values.input:
        parameter.value = value

    assert tuple(values.expected) == tuple(values.collected)


def test_postpopulated_connections(qtbot):
    values = epyqlib.tests.common.Values(
        initial=42,
        input=[42, 37, 23],
        expected=["37", "23"],
    )

    parameter = Parameter(
        name="Parameter A",
        value=values.initial,
    )

    root = Root()
    model = epyqlib.attrsmodel.Model(
        root=root,
        columns=columns,
    )

    root.append_child(parameter)

    data_changed = DataChangedCollector(
        model=model,
        parameter=parameter,
        column=1,
        roles=(PyQt5.QtCore.Qt.DisplayRole,),
    )

    model.model.dataChanged.connect(data_changed.collect)
    data_changed.collected.connect(values.collect)

    for value in values.input:
        parameter.value = value

    assert tuple(values.expected) == tuple(values.collected)


def test_with_pyqtpropertys(qtbot):
    @graham.schemify(tag="pyqtproperty_parameter")
    @epyqlib.attrsmodel.ify()
    @epyqlib.utils.qt.pyqtify(
        property_decorator=lambda: PyQt5.QtCore.pyqtProperty("PyQt_PyObject"),
    )
    @attr.s(hash=False)
    class PyQtPropertyParameter(epyqlib.treenode.TreeNode):
        type = attr.ib(default="test_parameter", init=False)
        name = attr.ib(default="New Parameter")
        value = attr.ib(
            default=None,
            converter=epyqlib.attrsmodel.to_decimal_or_none,
        )
        uuid = epyqlib.attrsmodel.attr_uuid()

        def __attrs_post_init__(self):
            super().__init__()

    Root = epyqlib.attrsmodel.Root(
        default_name="Parameters", valid_types=(PyQtPropertyParameter, Group)
    )

    columns = epyqlib.attrsmodel.columns(
        ((PyQtPropertyParameter, "name"),),
        ((PyQtPropertyParameter, "value"),),
    )

    model = make_a_model(
        root_cls=Root,
        parameter_cls=PyQtPropertyParameter,
        columns=columns,
    )

    parameter = node_from_name(model, "Parameter B")

    values = epyqlib.tests.common.Values(
        initial=42,
        input=[42, 37],
        expected=["37"],
    )

    parameter.value = values.initial

    data_changed = DataChangedCollector(
        model=model,
        parameter=parameter,
        column=1,
        roles=(PyQt5.QtCore.Qt.DisplayRole,),
    )

    model.model.dataChanged.connect(data_changed.collect)
    data_changed.collected.connect(values.collect)

    for value in values.input:
        parameter.value = value

    parameter.tree_parent.remove_child(child=parameter)

    parameter.value += 1

    assert tuple(values.expected) == tuple(values.collected)


def test_columns():
    columns = epyqlib.attrsmodel.columns(
        ((Parameter, "name"), (Group, "name")),
        ((Parameter, "value"),),
    )

    name_column = epyqlib.attrsmodel.Column(
        name="Name",
        fields={Parameter: "name", Group: "name"},
    )

    assert columns[0] == name_column
    # TODO: using the human name?  seems kinda bad in code
    assert columns["Name"] == name_column
    assert columns[Parameter, "name"] == name_column


def test_children_property_retained():
    @epyqlib.attrsmodel.ify()
    @epyqlib.utils.qt.pyqtify()
    @attr.s(hash=False)
    class N(epyqlib.treenode.TreeNode):
        children = attr.ib(default=attr.Factory(list), init=False)

        def __attrs_post_init__(self):
            super().__init__()

    n = N()

    assert isinstance(inspect.getattr_static(n, "children"), property)


def test_children_changed_signals():
    model = make_a_model()

    group = node_from_name(model, "Group C")

    added_items = []

    def added(item, row):
        added_items.append((item, row))

    removed_items = []

    def removed(parent, item, row):
        removed_items.append((parent, item, row))

    group.pyqt_signals.child_added.connect(added)
    group.pyqt_signals.child_removed.connect(removed)

    parameter = Parameter()
    group.append_child(parameter)
    assert added_items == [(parameter, 0)]

    group.remove_child(child=parameter)

    assert added_items == [(parameter, 0)]
    assert removed_items == [(group, parameter, 0)]


@graham.schemify(tag="pass_through")
@epyqlib.attrsmodel.ify()
@epyqlib.utils.qt.pyqtify()
@epyqlib.utils.qt.pyqtify_passthrough_properties(
    original="original",
    field_names=("value",),
)
@attr.s(hash=False)
class PassThrough(epyqlib.treenode.TreeNode):
    original = attr.ib(
        metadata=graham.create_metadata(
            field=epyqlib.attrsmodel.Reference(),
        ),
    )
    value = attr.ib(
        default=None,
        converter=epyqlib.attrsmodel.to_decimal_or_none,
    )
    uuid = epyqlib.attrsmodel.attr_uuid()

    def __attrs_post_init__(self):
        super().__init__()


PassThroughRoot = epyqlib.attrsmodel.Root(
    default_name="Pass Through", valid_types=(Parameter, PassThrough)
)


def test_original_signals(qtbot):
    root = PassThroughRoot()

    columns = epyqlib.attrsmodel.columns(
        ((Parameter, "name"),),
        (
            (Parameter, "value"),
            (PassThrough, "value"),
        ),
    )

    model = epyqlib.attrsmodel.Model(
        root=root,
        columns=columns,
    )

    original = Parameter()
    passthrough_a = PassThrough(original=None, value=None)
    passthrough_b = PassThrough(original=original, value=None)

    root.append_child(original)
    root.append_child(passthrough_a)
    root.append_child(passthrough_b)

    common_expected = ["5"]

    values = {
        "original": epyqlib.tests.common.Values(
            initial=None,
            input=None,
            expected=common_expected,
        ),
        "a": epyqlib.tests.common.Values(
            initial=None,
            input=None,
            expected=common_expected,
        ),
        "b": epyqlib.tests.common.Values(
            initial=None,
            input=None,
            expected=common_expected,
        ),
    }

    column = 1
    collectors = {
        "original": DataChangedCollector(
            model=model,
            parameter=original,
            column=column,
            roles=(PyQt5.QtCore.Qt.DisplayRole,),
        ),
        "a": DataChangedCollector(
            model=model,
            parameter=passthrough_a,
            column=column,
            roles=(PyQt5.QtCore.Qt.DisplayRole,),
        ),
        "b": DataChangedCollector(
            model=model,
            parameter=passthrough_b,
            column=column,
            roles=(PyQt5.QtCore.Qt.DisplayRole,),
        ),
    }

    for name in values:
        model.model.dataChanged.connect(collectors[name].collect)
        collectors[name].collected.connect(values[name].collect)

    passthrough_a.original = original

    assert passthrough_a.value == passthrough_b.value == original.value

    passthrough_a.value = 5

    assert passthrough_a.value == passthrough_b.value == original.value

    for name, value in values.items():
        assert tuple(value.expected) == tuple(value.collected)


def test_to_decimal_or_none_re_locale():
    decimal_string = "1,000.00"

    with epyqlib.tests.common.use_locale(("C", None)):
        with pytest.raises(ValueError, match=repr(decimal_string)):
            epyqlib.attrsmodel.to_decimal_or_none(decimal_string)

    with epyqlib.tests.common.use_locale(("en_US", "utf8"), "us"):
        epyqlib.attrsmodel.to_decimal_or_none(decimal_string)


def test_to_int_or_none_re_locale():
    int_string = "1,000"

    with epyqlib.tests.common.use_locale(("C", None)):
        with pytest.raises(ValueError, match=repr(int_string)):
            epyqlib.attrsmodel.to_int_or_none(int_string)

    with epyqlib.tests.common.use_locale(("en_US", "utf8"), "us"):
        epyqlib.attrsmodel.to_int_or_none(int_string)


def test_two_state_checkbox():
    @graham.schemify(tag="checkbox_parameter")
    @epyqlib.attrsmodel.ify()
    @epyqlib.utils.qt.pyqtify()
    @attr.s(hash=False)
    class CheckboxParameter(epyqlib.treenode.TreeNode):
        type = attr.ib(default="test_parameter", init=False)
        name = attr.ib(default="New Parameter")
        value = attr.ib(
            default=None,
            converter=epyqlib.attrsmodel.two_state_checkbox,
        )
        uuid = epyqlib.attrsmodel.attr_uuid()

        def __attrs_post_init__(self):
            super().__init__()

    Root = epyqlib.attrsmodel.Root(
        default_name="Parameters", valid_types=(CheckboxParameter, Group)
    )

    columns = epyqlib.attrsmodel.columns(
        (
            (CheckboxParameter, "name"),
            (Group, "name"),
        ),
        ((CheckboxParameter, "value"),),
    )

    root = Root()
    model = epyqlib.attrsmodel.Model(
        root=root,
        columns=columns,
    )

    parameter = CheckboxParameter()
    root.append_child(parameter)

    index = model.index_from_node(parameter)
    column_index = columns.index_of("Value")
    index = index.siblingAtColumn(column_index)

    flags = model.model.flags(index)
    assert flags & PyQt5.QtCore.Qt.ItemIsUserCheckable


def test_enumeration(qtbot):
    @graham.schemify(tag="leaf")
    @epyqlib.attrsmodel.ify()
    @epyqlib.utils.qt.pyqtify()
    @attr.s(hash=False)
    class TestEnumerationLeaf(epyqlib.treenode.TreeNode):
        name = attr.ib(default="New Leaf")
        enumeration_uuid = attr.ib(
            default=None,
            converter=epyqlib.attrsmodel.convert_uuid,
        )
        epyqlib.attrsmodel.attrib(
            attribute=enumeration_uuid,
            delegate=epyqlib.attrsmodel.RootDelegateCache(
                list_selection_root="test list_selection_root",
            ),
            converter=epyqlib.attrsmodel.convert_uuid,
        )
        uuid = epyqlib.attrsmodel.attr_uuid()

        def __attrs_post_init__(self):
            super().__init__()

    @graham.schemify(tag="group")
    @epyqlib.attrsmodel.ify()
    @epyqlib.utils.qt.pyqtify()
    @attr.s(hash=False)
    class TestEnumerationGroup(epyqlib.treenode.TreeNode):
        name = attr.ib(default="New Group")
        children = attr.ib(
            default=attr.Factory(list),
            cmp=False,
            init=False,
            metadata={"valid_types": (TestEnumerationLeaf,)},
        )
        uuid = epyqlib.attrsmodel.attr_uuid()

        def __attrs_post_init__(self):
            super().__init__()

        def can_drop_on(self, node):
            return isinstance(node, tuple(self.addable_types().values()))

    Root = epyqlib.attrsmodel.Root(
        default_name="Test", valid_types=(TestEnumerationLeaf, TestEnumerationGroup)
    )

    columns = epyqlib.attrsmodel.columns(
        (
            (TestEnumerationLeaf, "name"),
            (TestEnumerationGroup, "name"),
        ),
        ((TestEnumerationLeaf, "enumeration_uuid"),),
    )

    root = Root(uuid="b05c413f-215c-4376-a107-5bce992ed7a3")
    model = epyqlib.attrsmodel.Model(
        root=root,
        columns=columns,
    )
    model.add_drop_sources(model)

    item = TestEnumerationLeaf(
        name="Outside",
        uuid="cdedbbd2-c596-42cc-be45-7eb7953cc5ad",
    )
    root.append_child(item)

    group = TestEnumerationGroup(
        name="Enumerations",
        uuid="06c2a6ad-00b2-49ac-a836-057daa1ddc2f",
    )
    root.append_child(group)
    model.list_selection_roots["test list_selection_root"] = group

    enumerator_a = TestEnumerationLeaf(
        name="Inside A",
        uuid="1900f7e3-7230-40c1-9f5f-b838e2c33710",
    )
    group.append_child(enumerator_a)
    enumerator_b = TestEnumerationLeaf(
        name="Inside B",
        uuid="b9aeea0a-94ea-4fe6-a627-50caa942fbb5",
    )
    group.append_child(enumerator_b)
    enumerator_c = TestEnumerationLeaf(
        name="Inside C",
        uuid="a6a4e027-e128-4860-9f64-8be93708916e",
    )
    group.append_child(enumerator_c)

    view = PyQt5.QtWidgets.QTreeView()
    view.setItemDelegate(epyqlib.attrsmodel.create_delegate())
    view.setModel(model.model)

    target_index = model.index_from_node(item)
    target_index = target_index.siblingAtColumn(
        columns.index_of("Enumeration Uuid"),
    )

    item.enumeration_uuid = enumerator_a.uuid

    application = qt_api.QApplication.instance()

    for row, enumerator in enumerate(group.children):
        assert view.edit(
            target_index,
            PyQt5.QtWidgets.QAbstractItemView.AllEditTriggers,
            None,
        )
        (editor,) = view.findChildren(epyqlib.attrsmodel.CustomMulti)

        editor.setCurrentRow(row)

        PyQt5.QtCore.QCoreApplication.postEvent(
            editor,
            PyQt5.QtGui.QKeyEvent(
                PyQt5.QtCore.QEvent.KeyPress,
                PyQt5.QtCore.Qt.Key_Enter,
                PyQt5.QtCore.Qt.NoModifier,
            ),
        )

        # this is fun.  if you get weird issues try doing this more times
        for _ in range(3):
            application.processEvents()

        assert enumerator.uuid == item.enumeration_uuid


def test_all_selection_roots_avail():
    @graham.schemify(tag="parameter")
    @epyqlib.attrsmodel.ify()
    @epyqlib.utils.qt.pyqtify()
    @attr.s(hash=False)
    class C(epyqlib.treenode.TreeNode):
        a = attr.ib(default=None)
        b = attr.ib(default=None)
        c = attr.ib(default=None)
        epyqlib.attrsmodel.attrib(
            attribute=c,
            delegate=epyqlib.attrsmodel.RootDelegateCache(
                list_selection_root="c root",
            ),
        )
        d = attr.ib(default=None)
        epyqlib.attrsmodel.attrib(
            attribute=d,
            delegate=epyqlib.attrsmodel.RootDelegateCache(
                list_selection_root="d root",
            ),
        )

        def __attrs_post_init__(self):
            super().__init__()

    expected = {"c": "c root", "d": "d root"}
    assert epyqlib.attrsmodel.list_selection_roots(C) == expected


def test_types_list_selection_roots():
    @graham.schemify(tag="parameter")
    @epyqlib.attrsmodel.ify()
    @epyqlib.utils.qt.pyqtify()
    @attr.s(hash=False)
    class C(epyqlib.treenode.TreeNode):
        a = attr.ib(default=None)
        b = attr.ib(default=None)

        c = attr.ib(default=None)
        epyqlib.attrsmodel.attrib(
            attribute=c,
            delegate=epyqlib.attrsmodel.RootDelegateCache(
                list_selection_root="c root",
            ),
        )

        d = attr.ib(default=None)
        epyqlib.attrsmodel.attrib(
            attribute=d,
            delegate=epyqlib.attrsmodel.RootDelegateCache(
                list_selection_root="d root",
            ),
        )

        def __attrs_post_init__(self):
            super().__init__()

    @graham.schemify(tag="parameter")
    @epyqlib.attrsmodel.ify()
    @epyqlib.utils.qt.pyqtify()
    @attr.s(hash=False)
    class D(epyqlib.treenode.TreeNode):
        e = attr.ib(default=None)
        f = attr.ib(default=None)

        g = attr.ib(default=None)
        epyqlib.attrsmodel.attrib(
            attribute=g,
            delegate=epyqlib.attrsmodel.RootDelegateCache(
                list_selection_root="g root",
            ),
        )

        h = attr.ib(default=None)
        epyqlib.attrsmodel.attrib(
            attribute=h,
            delegate=epyqlib.attrsmodel.RootDelegateCache(
                list_selection_root="h root",
            ),
        )

        def __attrs_post_init__(self):
            super().__init__()

    types = epyqlib.attrsmodel.Types(types=(C, D))

    expected = {
        "c root",
        "d root",
        "g root",
        "h root",
    }

    assert set(types.list_selection_roots()) == expected


def test_noneditable_columns():
    model = make_a_model()

    group = model.root.children[0]
    assert isinstance(group, Group)

    base_index = model.index_from_node(group)

    index = base_index.siblingAtColumn(columns.index_of("Value"))
    item = model.model.itemFromIndex(index)
    assert not item.isEditable()


def test_editable_columns():
    model = make_a_model()

    group = model.root.children[0]
    assert isinstance(group, Group)

    base_index = model.index_from_node(group)

    index = base_index.siblingAtColumn(columns.index_of("Name"))
    item = model.model.itemFromIndex(index)
    assert item.isEditable()


def test_none_values_show_dash():
    model = make_a_model()

    group = model.root.children[0]
    assert isinstance(group, Group)

    base_index = model.index_from_node(group)
    index = base_index.siblingAtColumn(columns.index_of("Name"))
    item = model.model.itemFromIndex(index)

    group.name = "The Name"
    assert group.name == item.data(PyQt5.QtCore.Qt.DisplayRole)
    assert group.name == item.data(PyQt5.QtCore.Qt.EditRole)

    group.name = None
    # TODO: CAMPid 0794305784527546542452654254679680
    # The display role is supposed to be '-' for None but they can't be
    # different
    #
    # http://doc.qt.io/qt-5/qstandarditem.html#data
    #   The default implementation treats Qt::EditRole and Qt::DisplayRole
    #   as referring to the same data
    assert "" == item.data(PyQt5.QtCore.Qt.DisplayRole)
    assert "" == item.data(PyQt5.QtCore.Qt.EditRole)



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import functools
import math

from PyQt5 import QtWidgets
from PyQt5.QtWidgets import QAbstractSlider
from PyQt5.QtCore import pyqtSignal, pyqtSlot, QModelIndex

import epyqlib.listmenu
import epyqlib.listmenuview_ui
import epyqlib.utils.qt

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class ListMenuView(QtWidgets.QWidget):
    node_clicked = pyqtSignal(epyqlib.listmenu.Node)

    def __init__(self, parent=None, in_designer=False):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.listmenuview_ui.Ui_Form()
        self.ui.setupUi(self)

        scroll_bar = self.ui.list_view.verticalScrollBar()
        up = functools.partial(
            scroll_bar.triggerAction, QAbstractSlider.SliderPageStepSub
        )
        self.ui.up_button.clicked.connect(up)
        down = functools.partial(
            scroll_bar.triggerAction, QAbstractSlider.SliderPageStepAdd
        )
        self.ui.down_button.clicked.connect(down)

        self.ui.list_view.viewport().setAutoFillBackground(False)

        self.ui.list_view.clicked.connect(self.clicked)

    @pyqtSlot(QModelIndex)
    def clicked(self, index):
        if not index.isValid():
            return

        node = self.model.node_from_index(index)

        self.node_clicked.emit(node)

        # self.ui.module_to_nv_button.clicked.connect(self.module_to_nv)
        # self.ui.write_to_module_button.clicked.connect(self.write_to_module)
        # self.ui.read_from_module_button.clicked.connect(self.read_from_module)
        # self.ui.write_to_file_button.clicked.connect(self.write_to_file)
        # self.ui.read_from_file_button.clicked.connect(self.read_from_file)
        #
        # self.resize_columns = epyqlib.nv.Columns(
        #     name=True,
        #     value=True)

    def setModel(self, model):
        self.model = model
        self.node_clicked.connect(self.model.node_clicked)

        self.ui.list_view.setModel(self.model)

        self.ui.esc_button.clicked.connect(self.model.esc_clicked)
        self.root_changed(self.model.root)
        self.model.root_changed.connect(self.root_changed)

        # model.set_status_string.connect(self.set_status_string)

        # self.ui.module_to_nv.connect(model.module_to_nv)
        # self.ui.read_from_module.connect(model.read_from_module)
        # self.ui.write_to_module.connect(model.write_to_module)
        # self.ui.read_from_file.connect(model.read_from_file)
        # self.ui.write_to_file.connect(model.write_to_file)

        # self.ui.tree_view.header().setStretchLastSection(False)

        # for i in epyqlib.nv.Columns.indexes:
        #     if self.resize_columns[i]:
        #         self.ui.tree_view.header().setSectionResizeMode(
        #             i, QtWidgets.QHeaderView.ResizeToContents)
        # # TODO: would be nice to share between message and signal perhaps?
        # self.ui.tree_view.header().setSectionResizeMode(
        #     epyqlib.nv.Columns.indexes.value, QtWidgets.QHeaderView.Stretch)
        #
        # self.ui.tree_view.setItemDelegateForColumn(
        #     epyqlib.nv.Columns.indexes.value,
        #     epyqlib.delegates.Combo(model=model, parent=self))

    @pyqtSlot(epyqlib.listmenu.Node)
    def root_changed(self, node):
        self.ui.label.setText(node.fields.name)
        self.ui.list_view.scrollToTop()
        self.ui.esc_button.setDisabled(node.tree_parent is None)

    def set_padding(self, padding):
        self.setStyleSheet(
            """
            QListView::item
            {{
                padding: {}px;
            }}
        """.format(
                padding
            )
        )

    def update_calculated_layout(self):
        minimum_padding = 0
        self.set_padding(minimum_padding)

        view = self.ui.list_view
        model = self.model

        list = view.contentsRect().height()
        item = (
            view.itemDelegate()
            .sizeHint(view.viewOptions(), model.index_from_node(model.root))
            .height()
        )
        most_items = math.floor(list / item)
        remainder = list % item
        padding = minimum_padding + (remainder / most_items) / 2
        self.set_padding(padding)

        for button in [self.ui.down_button, self.ui.up_button, self.ui.esc_button]:
            # TODO: CAMPid 98754713241621231778985432
            button.setMaximumWidth(button.height())

    def selected_text(self):
        indexes = self.ui.list_view.selectedIndexes()
        index = indexes[0]
        return index.internalPointer().fields.name

    def select_node(self, node):
        self.ui.list_view.setCurrentIndex(self.model.index_from_node(node))


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import json
import os
from base64 import b64decode


def ensure_dir(dir_name: str):
    if os.path.exists(dir_name):
        if os.path.isdir(dir_name):
            return
        else:
            raise NotADirectoryError(
                f"Files cache dir {dir_name} already exists but is not a directory"
            )

    os.makedirs(dir_name, exist_ok=True)


def decode(s: str) -> str:
    # Add missing padding just in case
    return json.loads(b64decode(s + "==="))


def decode_jwt(jwt: str) -> dict:
    [header, payload, signature] = jwt.split(".")
    return {
        "header": decode(header),
        "payload": decode(payload),
        "signature": signature,
    }


# jwt = "eyJ0eXAiOiJKV1QiLCJhbGciOiJIUzI1NiJ9.eyJzdWIiOiIxMjM0NTY3ODkwIiwibmFtZSI6IkpvaG4gRG9lIiwiYWRtaW4iOnRydWUsImp0aSI6IjJiZTgxYmEwLThiMWYtNGMyMC04NWI2LWUwMGU0NThiNzM4OCIsImlhdCI6MTU1NDIzMDg4MCwiZXhwIjoxNTU0MjM0NDgwfQ.pzquCfp7SGuh9ZG3w-Opp6pCFCGIMshIh97yIB8oIow"
# print(json.dumps(decode_jwt(jwt), indent=4))



from epyqlib.tabs.files.aws_login_manager import AwsLoginManager
from epyqlib.tabs.files.sync_config import SyncConfig


class BucketManager:
    _bucket_names = {
        "internal": "epc-files-internal-203222013089",
        "client": "epc-files-client-203222013089",
    }
    _bucket_name = _bucket_names[SyncConfig.get_env()]
    _logs_path = "logs/"
    _tag = "[Bucket Manager]"

    def __init__(self):
        self._aws = AwsLoginManager.get_instance()
        self._uploaded_logs: set[str] = None

    async def download_file(self, hash: str, filename: str):
        return await self._download(filename, "files/" + hash)

    async def download_log(self, hash: str, filename: str):
        # Is it really worth it to have logs in a separate folder...?
        return await self._download(filename, "logs/" + hash)

    async def _download(self, filename: str, key: str):
        try:
            bucket = self._aws.get_s3_resource().Bucket(self._bucket_name)
            bucket.download_file(Filename=filename, Key=key)
            print(f"{self._tag} Finished downloading {key}")
        except Exception as ex:
            import sys

            error_message = (
                f"{self._tag} Error downloading Key: {key} from Bucket: {self._bucket_name}.\n"
                + f"{str(ex)}\n"
            )
            sys.stderr.write(error_message)
            raise Exception(error_message)

    async def upload_log(self, source_path: str, dest_filename: str):
        print(f"{self._tag} Starting to upload log {dest_filename}")

        # TODO: Figure out if logs should really be uploaded to their own folder
        with open(source_path, "rb") as source_file:
            s3_resource = self._aws.get_s3_resource()
            bucket = s3_resource.Bucket(self._bucket_name)
            bucket.put_object(Key=self._logs_path + dest_filename, Body=source_file)

        print(f"{self._tag} Finished upload of log {dest_filename}")
        if self._uploaded_logs is not None:
            self._uploaded_logs.add(
                dest_filename
            )  # Assuming dest_filename is the file's hash

    def fetch_uploaded_log_names(self):
        if self._uploaded_logs is None:
            s3_resource = self._aws.get_s3_resource()
            bucket = s3_resource.Bucket(self._bucket_name)

            # Get list of all paths and trim the leading "logs/" from each
            self._uploaded_logs = set(
                [o.key[5:] for o in bucket.objects.filter(Prefix="logs/")]
            )

        return self._uploaded_logs


if __name__ == "__main__":
    bucket_manager = BucketManager()
    bucket_manager.fetch_uploaded_log_names()
    print(bucket_manager._uploaded_logs)



import logging
from datetime import datetime, timedelta

import attr
import boto3
from boto3 import Session
from boto3_type_annotations.cognito_identity import Client as CognitoIdentityClient
from boto3_type_annotations.cognito_idp import Client as CognitoIdpClient
from boto3_type_annotations.s3 import ServiceResource as S3Resource
from epyqlib.tabs.files.files_utils import decode_jwt
from epyqlib.tabs.files.sync_config import SyncConfig, Vars
from epyqlib.utils.general import safe_get

logger = logging.getLogger("CognitoHelper")


@attr.s(slots=True, auto_attribs=True)
class CognitoException(Exception):
    message: str


class CognitoHelper:

    _tag = "[CognitoHelper]"

    configs = {
        "internal": {
            "identity_pool_id": "us-west-2:2e8a787f-e6b3-4fc0-b24f-2f7855b9bcf3",
            "client_id": "6ald2becsbriio982480bjacgq",
            "region": "us-west-2",
            "user_pool_id": "cognito-idp.us-west-2.amazonaws.com/us-west-2_wIVmyFHI5",
        },
        "client": {
            "identity_pool_id": "us-west-2:92cfea99-64ff-4c20-9202-b752d9e0db61",
            "client_id": "2jof2obe6dtofat1eml8kqjssv",
            "region": "us-west-2",
            "user_pool_id": "cognito-idp.us-west-2.amazonaws.com/us-west-2_DdfcP6fUX",
        },
    }

    def __init__(self, environment: str):
        self._access_token = ""
        self._expires_in = 0
        self._expires_time = datetime.min
        self._id_token = ""
        self._decoded_id_token: dict = {}
        self._refresh_token = ""
        self._token_type = ""
        self.config = self.configs[environment]

        self._s3_resource: S3Resource = None

        self._session: Session = boto3.session.Session(
            region_name=self.config["region"],
            aws_access_key_id="",
            aws_secret_access_key="",
            aws_session_token="",
        )

        self._sync_config = SyncConfig.get_instance()

    def authenticate(self, username: str, password: str):
        """
        :raises botocore.errorfactory.UserNotFoundException If username is wrong
        :raises botocore.errorfactory.NotAuthorizedException If password is wrong
        """
        print(f"{self._tag} Beginning authentication for user {username}")
        client: CognitoIdpClient = self._get_anonymous_client("cognito-idp")

        try:
            response = client.initiate_auth(
                ClientId=self.config["client_id"],
                AuthFlow="USER_PASSWORD_AUTH",
                AuthParameters={"USERNAME": username, "PASSWORD": password},
            )
        except Exception:
            raise CognitoException("Invalid credentials.")

        result = response["AuthenticationResult"]
        self._access_token = result["AccessToken"]
        self._expires_in = result["ExpiresIn"]
        self._expires_time = datetime.now() + timedelta(seconds=self._expires_in)
        self._id_token = result["IdToken"]
        self._decoded_id_token = decode_jwt(self._id_token)["payload"]
        self._refresh_token = result["RefreshToken"]
        self._token_type = result["TokenType"]

        self._sync_config.set(Vars.refresh_token, self._refresh_token)
        # auth result has the keys: AccessToken, ExpiresIn, IdToken, RefreshToken, TokenType

        self._init_auth_session()

        print(f"{self._tag} Authentication successful")
        self._create_resources()

    def _init_auth_session(self):
        cognito: CognitoIdentityClient = boto3.client(
            "cognito-identity", region_name=self.config["region"]
        )

        id = cognito.get_id(
            AccountId="203222013089",
            IdentityPoolId=self.config["identity_pool_id"],
            Logins={self.config["user_pool_id"]: self._id_token},
        )

        response = cognito.get_credentials_for_identity(
            IdentityId=id["IdentityId"],
            Logins={self.config["user_pool_id"]: self._id_token},
        )
        credentials = response["Credentials"]

        self._session = boto3.session.Session(
            region_name=self.config["region"],
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretKey"],
            aws_session_token=credentials["SessionToken"],
        )

    def _get_anonymous_client(self, type: str):
        return boto3.client(
            type,
            region_name=self.config["region"],
            aws_access_key_id="",
            aws_secret_access_key="",
            aws_session_token="",
        )

    def get_s3_resource(self):
        if not self.is_session_valid():
            self._refresh()

        return self._s3_resource

    def _create_resources(self):
        self._s3_resource = self._session.resource("s3")

    def _get_client(self, type: str):
        return self._session.client(type)

    def _get_resource(self, service_name: str):
        return self._session.resource(service_name)

    def is_session_valid(self):
        return datetime.now() < self._expires_time

    def _get_refresh_token_pref(self):
        return self._sync_config.get(Vars.refresh_token)

    def _clear_refresh_token_pref(self):
        self._sync_config.set(Vars.refresh_token, None)

    def is_user_logged_in(self) -> bool:
        token = self._get_refresh_token_pref()
        return token is not None and token != ""

    def get_user_customer(self) -> str:
        return self._decoded_id_token.get("custom:customer")

    def get_username(self) -> str:
        return self._decoded_id_token.get("cognito:username")

    def is_user_epc(self) -> bool:
        return self.get_user_customer() == "epc"

    def _refresh(self, refresh_token: str = None, force=False):
        if self.is_session_valid() and not force:
            # If we have credentials that haven't expired yet, bail out
            return

        refresh_token = refresh_token or self._get_refresh_token_pref()
        client: CognitoIdpClient = self._get_anonymous_client("cognito-idp")

        try:
            response = client.initiate_auth(
                ClientId=self.config["client_id"],
                AuthFlow="REFRESH_TOKEN",
                AuthParameters={
                    "REFRESH_TOKEN": refresh_token,
                },
            )
        except client.exceptions.NotAuthorizedException:
            print(
                f"{self._tag} Previous refresh token invalid. Clearing and forcing user to log in again."
            )
            self._clear_refresh_token_pref()
            return

        # auth result has the keys: AccessToken, ExpiresIn, IdToken, TokenType
        result = response["AuthenticationResult"]
        self._access_token = result["AccessToken"]
        self._expires_in = result["ExpiresIn"]
        self._expires_time = datetime.now() + timedelta(seconds=self._expires_in)
        self._id_token = result["IdToken"]
        self._decoded_id_token = decode_jwt(self._id_token)["payload"]
        self._token_type = result["TokenType"]

        self._init_auth_session()
        self._create_resources()

    def log_out(self):
        self._access_token = ""
        self._expires_in = ""
        self._expires_time = datetime.min
        self._id_token = ""
        self._decoded_id_token = ""
        self._token_type = ""

        self._s3_resource = None
        self._sync_config.set(Vars.refresh_token, None)


if __name__ == "__main__":
    # refresh_token = "..."
    # config = SyncConfig.get_instance()

    import sys

    env = sys.argv[1]
    username = sys.argv[2]
    password = sys.argv[3]

    helper = CognitoHelper(env)
    helper.authenticate(username, password)
    print(helper._id_token)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['epyqlib',
 'epyqlib.autodevice',
 'epyqlib.autodevice.template',
 'epyqlib.cli',
 'epyqlib.pm',
 'epyqlib.tabs',
 'epyqlib.tabs.files',
 'epyqlib.tests',
 'epyqlib.tests.autodevice',
 'epyqlib.tests.pm',
 'epyqlib.tests.sync',
 'epyqlib.tests.utils',
 'epyqlib.twisted',
 'epyqlib.utils',
 'epyqlib.widgets']

package_data = \
{'': ['*'], 'epyqlib': ['resources/*']}

install_requires = \
['GitPython==2.1.15',
 'Pint==0.11',
 'PyQt5>=5.13.0',
 'PyQt5Designer>=5.14.1,<6.0.0',
 'QtAwesome==0.6.0',
 'Twisted==21.2.0',
 'alqtendpy==0.0.4',
 'appdirs==1.4.3',
 'arrow==0.12.1',
 'attrs>=19.3.0',
 'bitstruct==6.0.0',
 'boto3-type-annotations==0.3.1',
 'boto3==1.14.41.0',
 'canmatrix==0.9.1',
 'certifi==2020.6.20',
 'click<=7.1.2',
 'dulwich==0.20.6',
 'fab==3.0.0',
 'graham==0.1.11',
 'marshmallow==2.16.3',
 'natsort==5.5.0',
 'paho-mqtt==1.4.0',
 'pyelftools==0.26',
 'pysunspec>=2.1.1,<3.0.0',
 'pytest-qt==3.3.0',
 'pytest-rerunfailures==5.0',
 'pytest-twisted<=1.13.4',
 'pytest-xvfb==1.2.0',
 'pytest==5.3.5',
 'python-can>=3.3.4,<4.0.0',
 'python-dateutil==2.7.5',
 'python-docx==0.8.7',
 'python-dotenv==0.9.1',
 'qt5reactor==0.5',
 'siphash-cffi>=0.1.4,<0.2.0',
 'sunspecdemo>=0.1.6,<0.2.0',
 'treq==21.1.0',
 'twine==1.13.0']

extras_require = \
{':sys_platform == "Darwin"': ['certitude==1.0.1']}

entry_points = \
{'console_scripts': ['autodevice = epyqlib.autodevice.cli:cli',
                     'buildui = buildui:compile_ui',
                     'cangenmanual = epyqlib.cangenmanual:_entry_point',
                     'collectdevices = epyqlib.collectdevices:main',
                     'contiguouscommits = '
                     'epyqlib.utils.contiguouscommits:_entry_point [dulwich]',
                     'epyqflash = epyqlib.flash:_entry_point',
                     'epyqlib = epyqlib.cli.main:cli',
                     'genbuildinfo = epyqlib.genbuildinfo:write_build_file',
                     'generateversion = epyqlib.cli.generateversion:cli',
                     'patchvenv = epyqlib.patchvenv:main',
                     'updateepc = epyqlib.updateepc:main',
                     'versionfile = epyqlib.cli.versionfile:cli'],
 'pytest11': ['epyqlib = epyqlib.tests.pytest_plugin']}

setup_kwargs = {
    'name': 'epyqlib',
    'version': '2022.2.1',
    'description': '',
    'long_description': None,
    'author': 'Alex Anker',
    'author_email': 'alex.anker@epcpower.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import attr

from epyqlib.tabs.files.activity_log import ActivityLog, Event
from epyqlib.tabs.files.graphql import API


@attr.s(slots=True, auto_attribs=True)
class ActivitySyncer:
    activityLog: ActivityLog
    api: API
    _syncing: bool = attr.ib(default=False)
    _is_offline: bool = attr.ib(default=False)

    def set_offline(self, is_offline: bool):
        self._is_offline = is_offline

    async def listener(self, _: Event):
        if self._is_offline is True:
            return

        if self._syncing is True:
            return

        self._syncing = True
        try:
            while self.activityLog.has_cached_events() is True:
                event: Event = self.activityLog.read_oldest_event()
                result = await self.api.create_activity(event)
                self.activityLog.remove(event)
        finally:
            self._syncing = False



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from PyQt5.QtCore import Qt, pyqtProperty
from PyQt5.QtGui import QColor

import epyqlib.widgets.abstractwidget
import epyqlib.mixins
import epyqlib.widgets.scale_ui


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class Scale(
    epyqlib.widgets.abstractwidget.AbstractWidget, epyqlib.mixins.OverrideRange
):
    def __init__(self, parent=None, in_designer=False):
        self.s_vertically_flipped = False

        # TODO: multiple inheritance *sigh* should get this for me eventually...
        self._min = 0
        self._max = 1

        self._breakpoints = [
            self._min + (self._max - self._min) * n for n in [0.10, 0.25, 0.75, 0.90]
        ]

        dark_red_transparent = QColor(Qt.darkRed)
        dark_red_transparent.setAlpha(0)
        dark_yellow_transparent = QColor(Qt.darkYellow)
        dark_yellow_transparent.setAlpha(0)
        dark_green_transparent = QColor(Qt.darkGreen)
        dark_green_transparent.setAlpha(0)

        self._colors = [
            dark_red_transparent,
            dark_yellow_transparent,
            dark_green_transparent,
            dark_yellow_transparent,
            dark_red_transparent,
        ]

        self._frame = None
        self._signal = None

        super().__init__(
            ui_class=epyqlib.widgets.scale_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

        self.update_configuration()

    @pyqtProperty(bool)
    def label_visible(self):
        return epyqlib.widgets.abstractwidget.AbstractWidget.label_visible.fget(self)

    @label_visible.setter
    def label_visible(self, new_visible):
        epyqlib.widgets.abstractwidget.AbstractWidget.label_visible.fset(
            self, new_visible
        )
        self.ui.units.setVisible(self.label_visible)
        self.update_metadata()

    @pyqtProperty(float)
    def lower_red_breakpoint(self):
        return self._breakpoints[0]

    @lower_red_breakpoint.setter
    def lower_red_breakpoint(self, breakpoint):
        self._breakpoints[0] = breakpoint
        self.update_configuration()

    @pyqtProperty(float)
    def lower_yellow_breakpoint(self):
        return self._breakpoints[1]

    @lower_yellow_breakpoint.setter
    def lower_yellow_breakpoint(self, breakpoint):
        self._breakpoints[1] = breakpoint
        self.update_configuration()

    @pyqtProperty(float)
    def upper_yellow_breakpoint(self):
        return self._breakpoints[2]

    @upper_yellow_breakpoint.setter
    def upper_yellow_breakpoint(self, breakpoint):
        self._breakpoints[2] = breakpoint
        self.update_configuration()

    @pyqtProperty(float)
    def upper_red_breakpoint(self):
        return self._breakpoints[3]

    @upper_red_breakpoint.setter
    def upper_red_breakpoint(self, breakpoint):
        self._breakpoints[3] = breakpoint
        self.update_configuration()

    @pyqtProperty(QColor)
    def lower_red_color(self):
        return self._colors[0]

    @lower_red_color.setter
    def lower_red_color(self, color):
        self._colors[0] = color
        self.update_configuration()

    @pyqtProperty(QColor)
    def lower_yellow_color(self):
        return self._colors[1]

    @lower_yellow_color.setter
    def lower_yellow_color(self, color):
        self._colors[1] = color
        self.update_configuration()

    @pyqtProperty(QColor)
    def green_color(self):
        return self._colors[2]

    @green_color.setter
    def green_color(self, color):
        self._colors[2] = color
        self.update_configuration()

    @pyqtProperty(QColor)
    def upper_yellow_color(self):
        return self._colors[3]

    @upper_yellow_color.setter
    def upper_yellow_color(self, color):
        self._colors[3] = color
        self.update_configuration()

    @pyqtProperty(QColor)
    def upper_red_color(self):
        return self._colors[4]

    @upper_red_color.setter
    def upper_red_color(self, color):
        self._colors[4] = color
        self.update_configuration()

    def set_value(self, value):
        self.ui.scale.setValue(value)

    def set_range(self, min=None, max=None):
        if self.override_range:
            min = self.minimum
            max = self.maximum

        self.ui.scale.setRange(min=float(min), max=float(max))

    def set_unit_text(self, units):
        self.ui.units.setText("[{}]".format(units))

    def update_configuration(self):
        try:
            self.ui.scale.setColorRanges(self._colors, self._breakpoints)
        except ValueError as e:
            print(e, file=sys.stderr)

        self.repaint()

    # s_flipped function allows for the qscale to be flipped or not when
    # in vertical orientation
    @pyqtProperty(bool)
    def s_flipped(self):
        return self.s_vertically_flipped

    @s_flipped.setter
    def s_flipped(self, value):
        self.s_vertically_flipped = value
        self.ui.scale.vertically_flipped = value


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

import epyqlib.twisted.cancalibrationprotocol as ccp
import argparse


class Frame:
    def __init__(self, index, time, type, id, data, comment):
        self.index = int(index)
        self.time = float(time)
        self.type = type
        self.id = (int(id, 16), len(id) > 3)
        self.data = [int(d, 16) for d in data]
        self.comment = comment

    @classmethod
    def len(cls):
        return 6

    def __str__(self):
        format = (
            "Index: {{}}, Time: {{:.4f}}, Type: {{}}, ID: {{:0{}X}}, "
            "Data: {{}}, Comment: {{}}"
        ).format(8 if self.id[1] else 3)
        data = ["{:02X}".format(d) for d in self.data]
        data = " ".join(data)
        return format.format(
            self.index, self.time, self.type, self.id[0], data, self.comment
        )


def load_trc(file):
    frame_length = Frame.len()

    for line in file:
        if line.lstrip().startswith(";"):
            continue

        elements = line.replace(")", "").split()
        data_length = int(elements[frame_length - 2])
        meta = elements[: frame_length - 2]
        data = elements[frame_length - 1 : frame_length + data_length - 1]
        try:
            comment = elements[frame_length + data_length - 1]
        except IndexError:
            comment = None

        frame = Frame(*meta, data, comment)
        if data_length != len(frame.data):
            raise Exception("blue")

        frame.time /= 1000

        yield frame


def load_sock(file):
    frame_length = Frame.len()

    for line in file:
        elements = line.replace("(", "").replace(")", "").replace("#", " ").split()
        id = elements[2]
        if id == "FFFFFFFF":
            comment = elements[3].lstrip("0123456789")
            data = elements[3][: -len(comment)]
        else:
            data = elements[3]
            data = [data[i : i + 2] for i in range(0, len(data), 2)]

        frame = Frame(
            index=0, time=elements[0], type="Rx", id=id, data=data, comment=None
        )

        yield frame


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("--trace", "-t", type=argparse.FileType("r"))
    parser.add_argument("--sock", "-s", type=argparse.FileType("r"))

    return parser.parse_args(args)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    args = parse_args(args=args)

    blocks = []

    if not ((args.trace is not None) ^ (args.sock is not None)):
        print("You must specify one and only one file type")
        return -1

    if args.trace:
        frames = load_trc(args.trace)
    elif args.sock:
        frames = load_sock(args.sock)

    for frame in frames:
        if frame.type == "Rx" and frame.id[1] and frame.id[0] == ccp.bootloader_can_id:
            if frame.data[0] == ccp.CommandCode.set_mta:
                address = sum(v << (8 * (3 - p)) for p, v in enumerate(frame.data[-4:]))
                # print('Raw Address: {}'.format(frame.data[-4:]))
                # print('Address: {}'.format(address))
                # if len(blocks) != 0:
                #     print('Latest block: {}'.format(blocks[-1]))
                #     print('Latest start: {}'.format(blocks[-1][0]))
                #     print('Latest length: {}'.format(blocks[-1][1]))
                if len(blocks) == 0 or blocks[-1][0] + blocks[-1][1] / 2 != address:
                    blocks.append([address, 0])
            elif frame.data[0] == ccp.CommandCode.download_6:
                blocks[-1][1] += 6
            elif frame.data[0] == ccp.CommandCode.download:
                blocks[-1][1] += frame.data[2]

    print("Note that separate blocks will be detected as one if they are contiguous.")
    for block in blocks:
        print(
            "Start: {:08X}, Length: {}, Bytes: {}".format(
                block[0], int(block[1] / 2), block[1]
            )
        )


if __name__ == "__main__":
    import sys

    sys.exit(main())



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import epyqlib.mixins
import epyqlib.widgets.abstracttxwidget
import epyqlib.widgets.horizontalslider_ui


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class HorizontalSlider(
    epyqlib.widgets.abstracttxwidget.AbstractTxWidget, epyqlib.mixins.OverrideRange
):
    def __init__(self, parent=None, in_designer=False):
        self._zero_count = None
        self._counts = 1000
        self._frame = None
        self._signal = None

        super().__init__(
            ui_class=epyqlib.widgets.horizontalslider_ui.Ui_Form,
            parent=parent,
            in_designer=in_designer,
        )

        self.ui.value.setTickInterval(self._counts / 4)
        self.ui.value.setRange(0, self._counts)

        self.ui.value.valueChanged.connect(self.widget_value_changed)

    def widget_value_changed(self, counts):
        value = counts - self._zero_count
        value *= (self._max - self._min) / self._counts

        epyqlib.widgets.abstracttxwidget.AbstractTxWidget.widget_value_changed(
            self, value
        )

    def set_value(self, value):
        if value is None:
            value = 0

        counts = (value * self._counts) / (self._max - self._min) + self._zero_count

        self.ui.value.setValue(int(round(counts)))

    # TODO: CAMPid 2397847962541678243196352195498
    def set_range(self, min=None, max=None):
        if self.override_range:
            min = self.minimum
            max = self.maximum

        if min == max:
            # TODO: pick the right exception
            raise Exception("Min and max may not be the same")
        elif min > max:
            # TODO: pick the right exception
            raise Exception("Min must be less than max")

        self._min = min
        self._max = max
        self._zero_count = (0 - min) / (max - min) * self._counts

        if self.signal_object is not None:
            min_string = self.signal_object.format_float(self._min)
            max_string = self.signal_object.format_float(self._max)
        else:
            min_string = ""
            max_string = ""

        self.ui.min.setText(min_string)
        self.ui.max.setText(max_string)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import collections
import enum
import json

import canmatrix.formats
import twisted.internet.defer

import PyQt5.QtWidgets

import epyqlib.canneo
import epyqlib.nv
import epyqlib.pm.valuesetmodel
import epyqlib.twisted.nvs
import epyqlib.utils.general
import epyqlib.utils.qt
import epyqlib.utils.twisted

__copyright__ = "Copyright 2017, EPC Power Corp."
__license__ = "GPLv2+"


class ConfigurationError(Exception):
    pass


class ValueTypes(enum.Enum):
    parameters = "auto_parameters"
    value_set = "auto_value_set"


def value_file(raw_dict):
    paths = {
        value_type: raw_dict[value_type.value]
        for value_type in ValueTypes
        if value_type.value in raw_dict
    }

    try:
        ((value_type, path),) = paths.items()
    except ValueError as e:
        raise ConfigurationError(
            "Expected one value file but got {}".format(paths),
        ) from e

    return value_type, path


def referenced_files(raw_dict):
    return value_file(raw_dict)[1:2]


class DeviceExtension:
    def __init__(self, device):
        self.device = device

        self.ui = None
        self.frames_nv = None
        self.nvs = None
        self.nv_protocol = None
        self.transport = None
        self.parameter_dict = None
        self.progress = None
        self.value_type = None
        self.parameter_names = None
        self.metas = None
        self.required_serial_number = None
        self.serial_number_names = None

    def post(self):
        self.value_type, parameter_path = value_file(self.device().raw_dict)
        parameter_path = self.device().absolute_path(parameter_path)

        self.ui = self.device().uis["Factory"]
        self.ui.load_parameters_button.clicked.connect(self.load_parameters)

        matrix_nv = list(canmatrix.formats.loadp(self.device().can_path).values())[0]
        self.frames_nv = epyqlib.canneo.Neo(
            matrix=matrix_nv,
            frame_class=epyqlib.nv.Frame,
            signal_class=epyqlib.nv.Nv,
            node_id_adjust=self.device().node_id_adjust,
            strip_summary=False,
        )

        self.required_serial_number = self.device().raw_dict["required_serial_number"]
        self.serial_number_names = self.device().raw_dict["serial_number_names"]

        access_level_path = self.device().raw_dict["access_level_path"]
        if access_level_path is not None:
            access_level_path = access_level_path.split(";")
        access_password_path = self.device().raw_dict["access_password_path"]
        if access_password_path is not None:
            access_password_path = access_password_path.split(";")

        # TODO: CAMPid 0794311304143707516085683164039671793972
        if self.device().raw_dict["nv_meta_enum"] == "Meta":
            self.metas = epyqlib.nv.meta_limits_first
        else:
            self.metas = (epyqlib.nv.MetaEnum.value,)

        self.nvs = epyqlib.nv.Nvs(
            neo=self.frames_nv,
            bus=self.device().bus,
            configuration=self.device().raw_dict["nv_configuration"],
            metas=self.metas,
            access_level_path=access_level_path,
            access_password_path=access_password_path,
        )
        self.nv_protocol = epyqlib.twisted.nvs.Protocol()
        from twisted.internet import reactor

        self.transport = epyqlib.twisted.busproxy.BusProxy(
            protocol=self.nv_protocol, reactor=reactor, bus=self.device().bus
        )

        if self.value_type == ValueTypes.parameters:
            with open(parameter_path, "r") as file:
                s = file.read()
                self.parameter_dict = json.loads(
                    s, object_pairs_hook=collections.OrderedDict
                )

            self.nvs.from_dict(self.parameter_dict)
            self.parameter_names = [k.split(":") for k in self.parameter_dict.keys()]
        elif self.value_type == ValueTypes.value_set:
            value_set = epyqlib.pm.valuesetmodel.loadp(parameter_path)
            self.nvs.from_value_set(value_set)
            self.parameter_names = [
                node.name.split(":") for node in value_set.model.root.leaves()
            ]

    @twisted.internet.defer.inlineCallbacks
    def no_gui_load_parameters(self):
        yield self._load_parameters()

    def load_parameters(self):
        d = self.no_gui_load_parameters()
        self._started()
        d.addBoth(epyqlib.utils.twisted.detour_result, self._ended)
        d.addCallback(epyqlib.utils.twisted.detour_result, self._finished)
        d.addErrback(epyqlib.utils.twisted.errbackhook)

        return d

    def _started(self):
        self.progress = epyqlib.utils.qt.Progress()
        self.progress.connect(
            progress=epyqlib.utils.qt.progress_dialog(parent=self.device().ui),
            label_text="Writing to device...",
        )

    def _ended(self):
        if self.progress is not None:
            self.progress.complete()
            self.progress = None

    def _finished(self):
        epyqlib.utils.qt.dialog(
            parent=self.device().ui,
            message=("Parameters successfully written to device and saved to NV"),
            icon=PyQt5.QtWidgets.QMessageBox.Information,
        )

    @twisted.internet.defer.inlineCallbacks
    def _check_serial_number(self):
        serial_nv = self.nvs.signal_from_names(
            "SN",
            "SerialNumber",
        )
        actual, _ = yield self.nv_protocol.read(
            nv_signal=serial_nv,
            meta=epyqlib.nv.MetaEnum.value,
        )
        actual = int(actual)
        expected = self.device().raw_dict["required_serial_number"]

        if actual != expected:
            raise epyqlib.utils.general.UnmatchedSerialNumberError(
                f"Expected {expected} but found {actual}"
            )

    @twisted.internet.defer.inlineCallbacks
    def _load_parameters(self):
        if self.required_serial_number is not None:
            yield self._check_serial_number()

        access_nodes = tuple(
            node
            for node in (
                self.nvs.password_node,
                self.nvs.access_level_node,
            )
            if node is not None
        )

        import attr

        def s(node, f):
            if f.name == "value":
                meta = node
            else:
                meta = getattr(node.meta, f.name)

            return str(meta.value)

        def p(nodes, i):
            print("writing {}: ".format(i))
            for node in nodes:
                print(
                    "    {} --- {}".format(
                        node,
                        ", ".join(s(node, f) for f in attr.fields(type(node.meta))),
                    )
                )

        p(access_nodes, 1)
        yield self.nvs.write_all_to_device(
            only_these=access_nodes,
            meta=(epyqlib.nv.MetaEnum.value,),
        )

        selected_nodes = set(self.nvs.all_nv()) - set(access_nodes)
        p(selected_nodes, 2)
        yield self.nvs.write_all_to_device(
            only_these=selected_nodes,
            meta=self.metas,
        )

        node_value_backups = []
        for node in access_nodes:
            node_value_backups.append((node, node.value))
            node.set_value(value=0)

        p(access_nodes, 3)
        yield self.nvs.write_all_to_device(
            only_these=access_nodes,
            meta=(epyqlib.nv.MetaEnum.value,),
        )

        for node, value in node_value_backups:
            node.set_value(value=value)

        yield self._module_to_nv()

    def _module_to_nv(self):
        self.nvs.save_signal.set_value(self.nvs.save_value)
        self.nvs.save_frame.update_from_signals()
        d = self.nv_protocol.write(
            nv_signal=self.nvs.save_signal,
            passive=True,
            meta=epyqlib.nv.MetaEnum.value,
        )
        d.addBoth(
            epyqlib.utils.twisted.detour_result,
            self._module_to_nv_off,
        )

        return d

    def _module_to_nv_off(self):
        self.nvs.save_signal.set_value(not self.nvs.save_value)
        d = self.nv_protocol.write(
            nv_signal=self.nvs.save_signal,
            passive=True,
            meta=epyqlib.nv.MetaEnum.value,
        )

        return d



import epyqlib.utils.general


def test_container():
    c = epyqlib.utils.general.Container(a=1, b=2)

    assert hasattr(c, "a")
    assert hasattr(c, "b")
    assert c.a == 1
    assert c.b == 2


def test_spaced_to_lower_camel():
    name = "ABC Def ghi LM1"
    expected = "abcDefGhiLM1"

    result = epyqlib.utils.general.spaced_to_lower_camel(name)

    assert result == expected


def test_spaced_to_upper_camel():
    name = "ABC Def ghi LM1"
    expected = "ABCDefGhiLM1"

    result = epyqlib.utils.general.spaced_to_upper_camel(name)

    assert result == expected


def test_underscored_to_upper_camel():
    name = "abc_de_fgh_IJK"
    expected = "AbcDeFghIjk"

    result = epyqlib.utils.general.underscored_to_upper_camel(name)

    assert result == expected


def test_underscored_camel_to_upper_camel():
    name = "abc_de_fgh_IJK"
    expected = "AbcDeFghIJK"

    result = epyqlib.utils.general.underscored_camel_to_upper_camel(name)

    assert result == expected


def test_cameled_to_spaced():
    name = "abcDeFghIjk - lmnOp"
    expected = "abc De Fgh Ijk - lmn Op"

    result = epyqlib.utils.general.cameled_to_spaced(name)

    assert result == expected


def test_cameled_to_spaced_dashed():
    name = "ABC - DefGhi"
    expected = "ABC - Def Ghi"

    result = epyqlib.utils.general.cameled_to_spaced(name)

    assert result == expected


def test_cameled_to_spaced_acronym():
    name = " XY1ABCDefGhiJKLMnOP1 "
    expected = "XY1 ABC Def Ghi JKL Mn OP1"

    result = epyqlib.utils.general.cameled_to_spaced(name)

    assert result == expected


def test_underscored_camel_to_title_spaced():
    name = "abcDeFghIjk_lmnOp"
    expected = "Abc De Fgh Ijk Lmn Op"

    result = epyqlib.utils.general.underscored_camel_to_title_spaced(name)

    assert result == expected


def test_underscored_camel_to_title_spaced_acronym():
    name = "XY1ABCDefGhi_lmnOP1"
    expected = "XY1 ABC Def Ghi Lmn OP1"

    result = epyqlib.utils.general.underscored_camel_to_title_spaced(name)

    assert result == expected


def test_underscored_camel_to_title_spaced_acronym_with_number_alone():
    name = "AB1"
    expected = "AB1"

    result = epyqlib.utils.general.underscored_camel_to_title_spaced(name)

    assert result == expected



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

from PyQt5.QtCore import pyqtProperty, QMarginsF, QSize
from PyQt5.QtSvg import QSvgWidget

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class SvgWidget(QSvgWidget):
    def __init__(self, parent=None, in_designer=False):
        QSvgWidget.__init__(self, parent=parent)

        self.in_designer = in_designer

        self._main_element = ""

    @pyqtProperty(str)
    def main_element(self):
        return self._main_element

    @main_element.setter
    def main_element(self, id):
        self._main_element = id

        self.update()

    def ratio(self):
        renderer = self.renderer()
        bounds = renderer.boundsOnElement(self.main_element)

        svg_dx = bounds.width()
        svg_dy = bounds.height()

        if svg_dx == 0:
            return 0

        return svg_dy / svg_dx

    def paintEvent(self, event):
        if len(self.main_element) > 0:
            renderer = self.renderer()
            bounds = renderer.boundsOnElement(self.main_element)
            svg_dx = bounds.width()
            svg_dy = bounds.height()
            svg_ratio = svg_dy / svg_dx

            widget_dx = self.width()
            widget_dy = self.height()
            widget_ratio = widget_dy / widget_dx

            if widget_ratio > svg_ratio:
                # widget is taller than svg item so pick taller view
                height = svg_dx * widget_ratio
                extra = (height - svg_dy) / 2
                margins = QMarginsF(0, extra, 0, extra)
            elif widget_ratio < svg_ratio:
                # widget is wider than svg item so pick wider view
                width = svg_dy / widget_ratio
                extra = (width - svg_dx) / 2
                margins = QMarginsF(extra, 0, extra, 0)
            else:
                margins = QMarginsF()

            bounds = bounds.marginsAdded(margins)

            renderer.setViewBox(bounds)

        QSvgWidget.paintEvent(self, event)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



#!/usr/bin/env python3

# TODO: """DocString if there is one"""

import functools

from PyQt5 import QtWidgets

import epyqlib.numberpad_ui
import epyqlib.utils.qt


# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class NumberPad(QtWidgets.QWidget):
    def __init__(self, parent=None, in_designer=False, action=None):
        super().__init__(parent=parent)

        self.in_designer = in_designer

        self.ui = epyqlib.numberpad_ui.Ui_Form()
        self.ui.setupUi(self)

        self.action = action
        self.ui.accept_button.clicked.connect(self.accept)
        self.ui.cancel_button.clicked.connect(functools.partial(self.exit, value=None))
        self.ui.up_button.clicked.connect(self.ui.cancel_button.clicked)

        numeric_buttons = []

        for number in range(10):
            button = getattr(self.ui, "button_{}".format(number))
            button.clicked.connect(functools.partial(self.ui.edit.insert, str(number)))
            numeric_buttons.append(button)

        self.ui.button_decimal.clicked.connect(
            functools.partial(self.ui.edit.insert, ".")
        )
        numeric_buttons.append(self.ui.button_decimal)

        self.ui.button_backspace.clicked.connect(self.ui.edit.backspace)

        for button in numeric_buttons:
            button.setStyleSheet(
                """
                QPushButton {
                    border: 0px;
                }
            """
            )

    def focus(self, value, action, label=""):
        self.ui.label.setText(label)
        self.ui.edit.setText(str(value))
        self.action = action
        self.parent().setCurrentWidget(self)
        focused_widget = self.focusWidget()
        if focused_widget is not None:
            focused_widget.clearFocus()

    def accept(self):
        try:
            value = float(self.ui.edit.text())
        except ValueError:
            value = None

        self.exit(value=value)

    def exit(self, value):
        self.action(value=value)


if __name__ == "__main__":
    import sys

    print("No script functionality here")
    sys.exit(1)  # non-zero is a failure



import functools
import epyqlib.numberpad
import epyqlib.abstractpluginclass

# See file COPYING in this source tree
__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


class NumberPadPlugin(epyqlib.abstractpluginclass.AbstractPlugin):
    def __init__(self, parent=None):
        epyqlib.abstractpluginclass.AbstractPlugin.__init__(self, parent=None)

        self._group = "EPC - General"
        self._init = epyqlib.numberpad.NumberPad
        self._module_path = "epyqlib.numberpad"
        self._name = "NumberPad"



import epyqlib.tabs.files.filesview
import epyqlib.abstractpluginclass

# See file COPYING in this source tree
__copyright__ = "Copyright 2018, EPC Power Corp."
__license__ = "GPLv2+"


# What does this do? When self_init was invalid value everything still worked as expected...
class FilesViewPlugin(epyqlib.abstractpluginclass.AbstractPlugin):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self._group = "EPC - General"
        self._init = epyqlib.tabs.files.filesview.FilesViewQtBuilder
        self._module_path = "epyqlib.filesview"
        self._name = "FilesView"



import datetime
import pathlib
import subprocess
import sys
import textwrap

import pkg_resources
import click


@click.command()
@click.option("--dist", "--distribution", "-d", "distribution", required=True)
@click.option("--root", "-r", default=None)
@click.option("--out", "-o", type=click.File("w"), default=sys.stdout)
def cli(distribution, root, out):
    version = pkg_resources.get_distribution(distribution).version

    sha = subprocess.check_output(
        [
            "git",
            "rev-parse",
            "--verify",
            "--quiet",
            "HEAD",
        ],
        cwd=root,
        encoding="utf-8",
    ).strip()

    actual_root_path = subprocess.check_output(
        [
            "git",
            "rev-parse",
            "--show-toplevel",
        ],
        cwd=root,
        encoding="utf-8",
    ).strip()

    file_path = pathlib.Path(__file__).relative_to(actual_root_path)

    contents = textwrap.dedent(
        f"""\
    # This file was automatically generated by {file_path}
    __version_build_time__ = '{datetime.datetime.utcnow().isoformat()}'
    
    __version__ = '{version}'
    __sha__ = '{sha}'
    __revision__ = '{sha}'
    """
    )

    out.write(contents)



import collections
import enum
import logging
import queue
import textwrap
import time

import attr
import twisted.internet.defer
import twisted.protocols.policies

import epyqlib.nv
import epyqlib.utils.general

__copyright__ = "Copyright 2016, EPC Power Corp."
__license__ = "GPLv2+"


logger = logging.getLogger(__name__)


class RequestTimeoutError(epyqlib.utils.general.ExpectedException):
    def __init__(self, state, item):
        message = "Protocol timed out while in state {} handling {}".format(
            state.name, item
        )

        super().__init__(message)
        self.state = state
        self.item = item

    def expected_message(self):
        return textwrap.dedent(
            """\
            Request timed out:
                {}, {}

            1. Confirm converter and adapter are on the same bus
            2. Confirm device file was loaded with node ID matching the converter's
            3. Possible parameter definition mismatch\
            """.format(
                self.state.name, self.item
            )
        )


class ReadOnlyError(Exception):
    pass


class SendFailedError(epyqlib.utils.general.ExpectedException):
    def expected_message(self):
        return "Send failed, make sure you are connected."


class CanceledError(Exception):
    pass


@enum.unique
class State(enum.Enum):
    idle = 0
    reading = 1
    writing = 2


@enum.unique
class Priority(enum.IntEnum):
    user = 0
    background = 1


@attr.s
class Request:
    priority = attr.ib()
    read = attr.ib(cmp=False)
    meta = attr.ib(cmp=False)
    signals = attr.ib(cmp=False)
    deferred = attr.ib(cmp=False)
    passive = attr.ib(cmp=False)
    all_values = attr.ib(cmp=False)
    frame = attr.ib(cmp=False)
    send_time = attr.ib(default=None)


class Protocol(twisted.protocols.policies.TimeoutMixin):
    def __init__(self, timeout=1):
        self._deferred = None

        self._state = State.idle
        self._previous_state = self._state

        self._active = False

        self._request_memory = None
        self._timeout = timeout

        self.requests = queue.PriorityQueue()

        self.cancel_queued = False

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        logger.debug("Entering state {}".format(new_state))
        self._previous_state = self._state
        self._state = new_state

    def makeConnection(self, transport):
        self._transport = transport
        logger.debug("Protocol.makeConnection(): {}".format(transport))

    def _start_transaction(self):
        if self._active:
            raise Exception("Protocol is already active")

        self._active = True

    def _transaction_over(self):
        self.setTimeout(None)
        import twisted.internet

        twisted.internet.reactor.callLater(0, self._transaction_over_after_delay)
        d = self._deferred
        self._deferred = None
        self._request_memory = None
        self.state = State.idle
        return d

    def _transaction_over_after_delay(self):
        self._active = False
        self._get()

    def read(
        self,
        nv_signal,
        meta,
        priority=Priority.background,
        passive=False,
        all_values=False,
    ):
        return self._read_write_request(
            nv_signals=(nv_signal,),
            read=True,
            meta=meta,
            priority=priority,
            passive=passive,
            all_values=all_values,
        )

    def read_multiple(
        self,
        nv_signals,
        meta,
        priority=Priority.background,
        passive=False,
        all_values=False,
    ):
        # TODO: make sure all signals are from the same frame
        return self._read_write_request(
            nv_signals=nv_signals,
            read=True,
            meta=meta,
            priority=priority,
            passive=passive,
            all_values=all_values,
        )

    def write(
        self,
        nv_signal,
        meta,
        priority=Priority.background,
        passive=False,
        ignore_read_only=False,
        all_values=False,
    ):
        # TODO: make sure all signals are from the same frame
        if nv_signal.frame.read_write.min > 0:
            if ignore_read_only:
                return
            else:
                raise ReadOnlyError()

        return self._read_write_request(
            nv_signals=(nv_signal,),
            read=False,
            meta=meta,
            priority=priority,
            passive=passive,
            all_values=all_values,
        )

    def write_multiple(
        self,
        nv_signals,
        meta,
        priority=Priority.background,
        passive=False,
        ignore_read_only=False,
        all_values=False,
    ):
        if tuple(nv_signals)[0].frame.read_write.min > 0:
            if ignore_read_only:
                return
            else:
                raise ReadOnlyError()

        if not isinstance(nv_signals, dict):
            nv_signals = {
                s: (
                    s.value
                    if meta == epyqlib.nv.MetaEnum.value
                    else getattr(s.meta, meta.name).value
                )
                for s in nv_signals
            }

        return self._read_write_request(
            nv_signals=nv_signals,
            read=False,
            meta=meta,
            priority=priority,
            passive=passive,
            all_values=all_values,
        )

    def _read_write_request(
        self, nv_signals, read, meta, priority, passive, all_values
    ):
        deferred = twisted.internet.defer.Deferred()

        if not isinstance(nv_signals, dict):
            nv_signals = {
                s: (
                    s.value
                    if meta == epyqlib.nv.MetaEnum.value
                    else getattr(s.meta, meta.name).value
                )
                for s in nv_signals
            }

        frame = tuple(nv_signals.keys())[0].frame

        self._put(
            Request(
                read=read,
                meta=meta,
                signals=nv_signals,
                deferred=deferred,
                priority=priority,
                passive=passive,
                all_values=all_values,
                frame=frame,
            )
        )

        return deferred

    def _put(self, request):
        self.requests.put(request)
        self._get()

    def _get(self):
        if not self._active:
            while True:
                try:
                    request = self.requests.get(block=False)
                except queue.Empty:
                    self.cancel_queued = False
                else:
                    if self.cancel_queued:
                        request.deferred.errback(CanceledError())
                    elif request.read:
                        self._read_write(request)
                    else:
                        self._read_before_write(request)

                if not self.cancel_queued:
                    break

    def _read_before_write(self, request):
        if isinstance(request.signals, dict):
            nonskip = request.signals
        else:
            nonskip = {
                s: (
                    s.value
                    if request.meta == epyqlib.nv.MetaEnum.value
                    else getattr(s.meta, request.meta.name).value
                )
                for s in request.signals
            }

        skip_signals = set(request.frame.parameter_signals) - set(nonskip.keys())

        if len(skip_signals) == 0:
            try:
                self._read_write(request)
            except Exception as e:
                request.deferred.errback(e)
        else:
            d = twisted.internet.defer.Deferred()
            d.callback(None)

            def read_then_write(args, nonskip=nonskip):
                values, meta = args
                data = {k: v for k, v in nonskip.items()}
                for s, v in values.items():
                    s = s.set_signal
                    if s in request.frame.parameter_signals:
                        if s not in data:
                            data[s] = s.from_human(v)

                return self.write_multiple(
                    nv_signals=data,
                    meta=request.meta,
                    all_values=True,
                )

            def write_response(args, nonskip=nonskip, request=request):
                values, meta = args
                data = {
                    signal.status_signal: values[signal.status_signal]
                    for signal in nonskip
                }

                request.deferred.callback((data, request.meta))

            d.addCallback(
                lambda _: self.read_multiple(
                    request.frame.parameter_signals, meta=request.meta, all_values=True
                )
            )
            d.addCallback(read_then_write)
            d.addCallback(write_response)
            d.addErrback(lambda e: request.deferred.errback(e))

    def _read_write(self, request):
        self._deferred = request.deferred
        try:
            self._start_transaction()
            self.state = State.reading if request.read else State.writing

            (read_write,) = (
                k
                for k, v in request.frame.read_write.enumeration.items()
                if v == ("Read" if request.read else "Write")
            )

            data = collections.OrderedDict()
            for signal in request.frame.signals:
                if signal is request.frame.read_write:
                    data[signal] = read_write
                elif signal.enumeration_name == "Meta":
                    data[signal] = request.meta.value
                elif signal not in request.frame.parameter_signals:
                    data[signal] = signal.value
                elif signal in request.signals and not request.read:
                    data[signal] = request.signals[signal]
                else:
                    data[signal] = None

                if data[signal] is None:
                    v = next(
                        v
                        for v in (
                            signal.from_human(signal.default_value),
                            signal.from_human(signal.min),
                            signal.from_human(signal.max),
                            0,
                        )
                        if v is not None
                    )
                    data[signal] = v

                data[signal] = int(data[signal])

            data = request.frame.update_from_signals(
                data=data.values(),
                only_return=True,
            )

            if request.passive:
                write = self._transport.write_passive
            else:
                write = self._transport.write

            self._request_memory = request

            if not write(request.frame.to_message(data)):
                self.send_failed()
                return

            request.send_time = time.time()

            self.setTimeout(self._timeout)
        except Exception as e:
            self.errback(e)

    def dataReceived(self, msg):

        if not self._active:
            return

        if self._deferred is None:
            return

        request = self._request_memory
        if request is None:
            return

        if not (
            msg.arbitration_id == request.frame.status_frame.id
            and (bool(msg.is_extended_id) == request.frame.status_frame.extended)
        ):
            return

        status_signal = tuple(request.signals)[0].status_signal

        if status_signal is None:
            return

        signals = status_signal.frame.unpack(msg.data, only_return=True)

        mux = status_signal.set_signal.frame.mux.value
        (response_mux_value,) = (
            v for k, v in signals.items() if k.name.endswith("_MUX")
        )
        if response_mux_value != mux:
            return
        meta_mux_value = tuple(
            v for k, v in signals.items() if k.enumeration_name == "Meta"
        )
        if len(meta_mux_value) == 1:
            (meta_mux_value,) = meta_mux_value
            if meta_mux_value != request.meta.value:
                print(" -- skipping due to unmatched meta")
                return

        response_read_write_value = signals[status_signal.frame.command_signal]
        # TODO: handle the enumeration
        if response_read_write_value != request.read:
            return

        self.setTimeout(None)

        if request.all_values:
            status_signals = {s.status_signal for s in request.signals}
            value = {
                s: s.to_human(value=v)
                for s, v in signals.items()
                if s in status_signals
            }
        else:
            raw_value = signals[status_signal]
            value = status_signal.to_human(value=raw_value)

        self.callback(value, request.meta)

    def send_failed(self):
        self.cancel_queued = True
        deferred = self._transaction_over()
        deferred.errback(SendFailedError())

    def timeoutConnection(self):
        request = self._request_memory
        # TODO: report all requested signals
        signal = tuple(request.signals)[0]
        mux_name = signal.frame.mux_name

        e = RequestTimeoutError(
            state=self.state,
            item=(
                f"{mux_name}:{signal.name} "
                f"({request.meta.name}, {request.send_time}, {time.time()}"
            ),
        )

        logger.debug(str(e))
        if self._previous_state in [State.idle]:
            self.state = self._previous_state
        deferred = self._transaction_over()
        deferred.errback(e)

    def callback(self, payload, meta):
        deferred = self._transaction_over()
        logger.debug("calling back for {}".format(deferred))
        deferred.callback((payload, meta))

    def errback(self, payload):
        deferred = self._transaction_over()
        logger.debug("erring back for {}".format(deferred))
        logger.debug("with payload {}".format(payload))
        deferred.errback(payload)

    def cancel(self):
        deferred = self._transaction_over()
        deferred.cancel()



# See file COPYING in this source tree
__copyright__ = "\n".join(
    [
        "Copyright (c) Twisted Matrix Laboratories",
        # https://github.com/twisted/twisted/blob/4368c0b84b82f0791f6df52dc80328f7bd493547/src/twisted/internet/_posixserialport.py
        "Copyright 2016, EPC Power Corp.",
    ]
)
__license__ = "GPLv2+"


import epyqlib.canneo


class BusProxy(epyqlib.canneo.QtCanListener):
    def __init__(self, protocol, reactor, bus=None, parent=None):
        super().__init__(receiver=self.readEvent, parent=parent)

        self._bus = bus
        self._reactor = reactor
        self._protocol = protocol

        if self._bus is not None:
            self.set_bus(bus=self._bus)

        self._protocol.makeConnection(self)
        # self.startReading()

    def set_bus(self, bus):
        bus.notifier.add(self)

        self._bus = bus

    def write(self, message):
        return self._bus.send(msg=message)

    def write_passive(self, message):
        return self._bus.send_passive(msg=message)

    def readEvent(self, message):
        """
        Some data's readable from serial device.
        """
        return self._protocol.dataReceived(message)

    def connectionLost(self, reason):
        """
        Called when the serial port disconnects.
        Will call C{connectionLost} on the protocol that is handling the
        serial data.
        """
        abstract.FileDescriptor.connectionLost(self, reason)
        self._serial.close()
        self.protocol.connectionLost(reason)



#!/usr/bin/env python3

import argparse
import collections
import io
import os
import platform
import requests
import shutil
import sys
import tempfile
import textwrap
import zipfile


Activate = collections.namedtuple(
    "Activate",
    [
        "windows",
        "set_format",
        "path_separator",
        "pwd",
        "comment_marker",
        "insert_before",
    ],
)

activate_scripts = {
    "activate": Activate(
        windows=False,
        set_format="export {name}={value}",
        path_separator=":",
        pwd="$(pwd)",
        comment_marker="#",
        insert_before=None,
    ),
    "activate.bat": Activate(
        windows=True,
        set_format="set {name}={value}",
        path_separator=";",
        pwd="%cd%",
        comment_marker="REM",
        insert_before=None,
    ),
    "activate.ps1": Activate(
        windows=True,
        set_format='${name} = "{value}"',
        path_separator=";",
        pwd="$((Get-Location).Path)",
        comment_marker="#",
        insert_before="# SIG",
    ),
}


def download_zips(directory):
    src_zips = {
        "fontawesome": "https://github.com/FortAwesome/Font-Awesome/archive/v4.6.3.zip",
        "metropolis": "https://github.com/chrismsimpson/Metropolis/archive/16882c2c2cb58405fd6a7d6a932a1dfc573b6813.zip",
    }

    os.makedirs(directory, exist_ok=True)

    for name, url in src_zips.items():
        response = requests.get(url)

        zip_data = io.BytesIO()
        zip_data.write(response.content)
        zip_file = zipfile.ZipFile(zip_data)
        zip_dir = os.path.split(zip_file.namelist()[0])[0]

        with tempfile.TemporaryDirectory() as td:
            zip_file.extractall(path=td)

            zip_path = os.path.join(td, zip_dir)

            destination = os.path.join(directory, name)
            if os.path.exists(destination):
                raise FileExistsError(
                    "`{}` already exists while extracting Zip".format(destination)
                )

            shutil.move(zip_path, destination)


def write_patch_notice(file, comment_marker):
    file.write(
        textwrap.dedent(
            """\


    {c}
    {c} ==== Patch below added to support EPyQ

    """.format(
                c=comment_marker
            )
        )
    )


def patch_activate(bin):
    path_variables = collections.OrderedDict(
        [
            (
                "PYQTDESIGNERPATH",
                [
                    ["sub", "epyqlib", "epyqlib"],
                    ["sub", "epyqlib", "epyqlib", "widgets"],
                ],
            )
        ]
    )

    for name, script in activate_scripts.items():
        path = os.path.join(bin, name)

        if os.path.exists(path):
            set_commands = []

            for variable, paths in path_variables.items():
                paths = [os.path.join(script.pwd, *p) for p in paths]

                command = script.set_format.format(
                    name=variable, value=script.path_separator.join(paths)
                )
                set_commands.append(command)

            if script.insert_before is not None:
                with open(path, "r") as f:
                    contents = f.read()
                with open(path, "w") as f:
                    written = False
                    for line in contents.splitlines():
                        if line.startswith(script.insert_before) and not written:
                            written = True
                            write_patch_notice(f, script.comment_marker)

                            for command in set_commands:
                                f.write(command + "\n")
                            f.write("\n\n")

                        f.write(line + "\n")
            else:
                with open(path, "a") as f:
                    write_patch_notice(f, script.comment_marker)

                    for command in set_commands:
                        f.write(command + "\n")


def write_activate_shortcuts(root, bin):
    for name, script in activate_scripts.items():
        caller = os.path.join(root, name)
        target = os.path.join(bin, name)
        if os.path.isfile(target):
            with open(caller, "w") as f:
                if script.windows:
                    target = '"{}"'.format(target.replace("\\", "/"))
                else:
                    target = 'source "{}"'.format(target)
                f.write(target + "\n")


def copy_designer_files(root):
    files = ["designer", "designer.bat", "designer.ps1", "designer.vbs"]

    for file in files:
        shutil.copy(
            os.path.join(os.path.dirname(__file__), "..", file),
            os.path.join(root, file),
        )


def parse_args(args):
    parser = argparse.ArgumentParser()
    parser.add_argument("--source", "--src", "-s", required=True)
    parser.add_argument("--bin", "-b", required=True)
    parser.add_argument("--root", "-r", required=True)
    parser.add_argument("--for-test", action="store_true")

    return parser.parse_args(args)


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    args = parse_args(args=args)

    download_zips(directory=args.source)
    patch_activate(bin=args.bin)
    if not args.for_test:
        write_activate_shortcuts(root=args.root, bin=args.bin)
        copy_designer_files(root=args.root)
        if platform.system() != "Windows":
            backup_cwd = os.getcwd()
            os.chdir(os.path.dirname(args.bin))
            os.symlink(os.path.basename(args.bin), "Scripts")
            os.chdir(backup_cwd)


if __name__ == "__main__":
    sys.exit(main())



import io
import textwrap

import pytest
from PyQt5 import QtWidgets
import PyQt5.uic

import epyqlib.tabs.files.filesview


def test_init(qtbot):
    view = epyqlib.tabs.files.filesview.FilesView()
    qtbot.add_widget(view)

    view.show()
    qtbot.wait_for_window_shown(widget=view)

    assert view.parent() is None
    assert isinstance(view.ui, epyqlib.tabs.files.filesview.Ui)
    assert not hasattr(view.ui, "something")


def test_build(qtbot):
    view = epyqlib.tabs.files.filesview.FilesView.build()
    qtbot.add_widget(view)

    view.show()
    qtbot.wait_for_window_shown(widget=view)

    assert view.parent() is None
    assert isinstance(view.ui, epyqlib.tabs.files.filesview.Ui)
    assert hasattr(view.ui, "files_grid")
